"""Student's t-distribution Mixture Model."""

# Author: Omri Tomer <omritomer1@mail.tau.ac.il>
# License: BSD 3 clause

import numpy as np

from scipy import linalg

from sklearn.mixture import GaussianMixture
from sklearn.utils.validation import check_is_fitted

import warnings

from sklearn import cluster
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils import check_random_state
from sklearn.utils.fixes import logsumexp

from .student_mixture_functions import _check_weights, _check_X, _check_locations, _check_dofs, _check_precisions,\
    _estimate_student_parameters, _compute_precision_cholesky, _compute_gamma_priors, _estimate_log_student_prob,\
    _initialize_dofs, _estimate_dofs
from .multivariate_t_functions import _multivariate_t_random, _multivariate_t_cdf


class StudentMixture(GaussianMixture):
    """Student Mixture.

    Representation of a Student's t- mixture model probability distribution.
    This class allows to estimate the parameters of a Student's mixture
    distribution.

    Parameters
    ----------
    n_components : int, defaults to 1.
        The number of mixture components.

    scale_type : {'full' (default), 'tied', 'diag', 'spherical'}
        String describing the type of scale parameters to use.
        Must be one of:

        'full'
            each component has its own general scale matrix
        'tied'
            all components share the same general scale matrix
        'diag'
            each component has its own diagonal scale matrix
        'spherical'
            each component has its own single variance

    algorithm : {'em' (default), 'mcecm'}
        String describing the algorithm used for estimating
        the degrees-of-freedom.
        Must be one of:

        'em'
            Expectation-Maximization algorithm
        'mcecm'
            Multicycle Expectation-Conditional-Maximization algorithm

    tol : float, defaults to 1e-3.
        The convergence threshold. EM iterations will stop when the
        lower bound average gain is below this threshold.

    reg_scale : float, defaults to 1e-6.
        Non-negative regularization added to the diagonal of scale.
        Allows to assure that the scale matrices are all positive.

    max_iter : int, defaults to 100.
        The number of EM iterations to perform.

    n_init : int, defaults to 1.
        The number of initializations to perform. The best results are kept.

    init_params : {'kmeans', 'random'}, defaults to 'kmeans'.
        The method used to initialize the weights, the locations and the
        precisions.
        Must be one of::

            'kmeans' : responsibilities are initialized using k-means.
            'random' : responsibilities are initialized randomly.

    dof_tol : float, defaults to 1e-3.
        The degrees-of-freedom estimation convergence threshold.

    dof_max_iter : int, defaults to 100.
        The number of iterations to perform for degrees-of-freedom estimation.

    weights_init : array-like, shape (n_components, ), optional
        The user-provided initial weights, defaults to None.
        If it None, weights are initialized using the `init_params` method.

    locations_init : array-like, shape (n_components, n_features), optional
        The user-provided initial locations, defaults to None,
        If it None, locations are initialized using the `init_params` method.

    precisions_init : array-like, optional.
        The user-provided initial precisions (inverse of the scale
        matrices), defaults to None.
        If it None, precisions are initialized using the 'init_params' method.
        The shape depends on 'scale_type'::

            (n_components,)                        if 'spherical',
            (n_features, n_features)               if 'tied',
            (n_components, n_features)             if 'diag',
            (n_components, n_features, n_features) if 'full'

    dofs_init : array-like, shape (n_components, ), optional
        The user-provided initial degrees-of-freedom, defaults to None (2 <= dofs_init[i] < inf for all i).
        If it None, dofs are initialized using the `init_params` method.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    warm_start : bool, default to False.
        If 'warm_start' is True, the solution of the last fitting is used as
        initialization for the next call of fit(). This can speed up
        convergence when fit is called several times on similar problems.
        In that case, 'n_init' is ignored and only a single initialization
        occurs upon the first call.
        See :term:`the Glossary <warm_start>`.

    verbose : int, default to 0.
        Enable verbose output. If 1 then it prints the current
        initialization and each iteration step. If greater than 1 then
        it prints also the log probability and the time needed
        for each step.

    verbose_interval : int, default to 10.
        Number of iteration done before the next print.

    Attributes
    ----------
    weights_ : array-like, shape (n_components,)
        The weights of each mixture components.

    locations_ : array-like, shape (n_components, n_features)
        The location of each mixture component.

    scales_ : array-like
        The scale of each mixture component.
        The shape depends on `scale_type`::

            (n_components,)                        if 'spherical',
            (n_features, n_features)               if 'tied',
            (n_components, n_features)             if 'diag',
            (n_components, n_features, n_features) if 'full'

    dofs_ : array-like, shape (n_components,)
        The degrees-of-freedom of each mixture components.

    precisions_ : array-like
        The precision matrices for each component in the mixture. A precision
        matrix is the inverse of a scale matrix. A scale matrix is
        symmetric positive definite so the mixture of Student's t-distributions
        can be equivalently parameterized by the precision matrices. Storing
        the precision matrices instead of the scale matrices makes it more
        efficient to compute the log-likelihood of new samples at test time.
        The shape depends on `scale_type`::

            (n_components,)                        if 'spherical',
            (n_features, n_features)               if 'tied',
            (n_components, n_features)             if 'diag',
            (n_components, n_features, n_features) if 'full'

    precisions_cholesky_ : array-like
        The cholesky decomposition of the precision matrices of each mixture
        component. A precision matrix is the inverse of a scale matrix.
        A scale matrix is symmetric positive definite so the mixture of
        Student's t-distributions can be equivalently parameterized by the
        precision matrices.
        Storing the precision matrices instead of the scale matrices makes
        it more efficient to compute the log-likelihood of new samples at test
        time. The shape depends on `scale_type`::

            (n_components,)                        if 'spherical',
            (n_features, n_features)               if 'tied',
            (n_components, n_features)             if 'diag',
            (n_components, n_features, n_features) if 'full'

    converged_ : bool
        True when convergence was reached in fit(), False otherwise.

    n_iter_ : int
        Number of step used by the best fit of EM to reach the convergence.

    lower_bound_ : float
        Lower bound value on the log-likelihood (of the training data with
        respect to the model) of the best fit of EM.


    """

    def __init__(self, n_components=1, scale_type='full',
                 algorithm='em', fixed_dofs=False,
                 tol=1e-3, reg_scale=1e-6,
                 max_iter=100, n_init=1, init_params='kmeans',
                 dof_tol=1e-3, dof_max_iter=100,
                 weights_init=None, locations_init=None, precisions_init=None, dofs_init=None,
                 random_state=None, warm_start=False,
                 verbose=0, verbose_interval=10):
        super().__init__(
            n_components=n_components, tol=tol, reg_covar=reg_scale,
            max_iter=max_iter, n_init=n_init, init_params=init_params,
            random_state=random_state, warm_start=warm_start,
            verbose=verbose, verbose_interval=verbose_interval)

        self.scale_type = scale_type
        self.algorithm = algorithm
        self.reg_scale = reg_scale
        self.fixed_dofs = fixed_dofs
        self.dof_tol = dof_tol
        self.dof_max_iter = dof_max_iter
        self.weights_init = weights_init
        self.locations_init = locations_init
        self.precisions_init = precisions_init
        self.dofs_init = dofs_init

    def _check_parameters(self, X):
        """Check the Student mixture parameters are well defined."""
        _, n_features = X.shape
        if self.scale_type not in ['spherical', 'tied', 'diag', 'full']:
            raise ValueError("Invalid value for 'scale_type': %s "
                             "'scale_type' should be in "
                             "['spherical', 'tied', 'diag', 'full']"
                             % self.scale_type)

        if self.weights_init is not None:
            self.weights_init = _check_weights(self.weights_init,
                                               self.n_components)

        if self.locations_init is not None:
            self.locations_init = _check_locations(self.locations_init,
                                                   self.n_components,
                                                   n_features)

        if self.precisions_init is not None:
            self.precisions_init = _check_precisions(self.precisions_init,
                                                     self.scale_type,
                                                     self.n_components,
                                                     n_features)

        if self.dofs_init is not None:
            self.dofs_init = _check_dofs(self.dofs_init,
                                         self.n_components)

    def _check_initial_parameters(self, X):
        """Check values of the basic parameters.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        """
        if self.reg_scale < 0.:
            raise ValueError("Invalid value for 'reg_scale': %.5f "
                             "regularization on scale must be "
                             "non-negative"
                             % self.reg_scale)
                             
        super()._check_initial_parameters(X)

        if self.algorithm not in ['em', 'mcecm']:
            raise ValueError("Invalid value for 'algorithm': %s "
                             "'algorithm' should be in "
                             "['em', 'mcecm']"
                             % self.algorithm)

        if not isinstance(self.fixed_dofs, bool):
            raise TypeError("Invalid value for 'fixed_dofs': %s "
                            "'fixed_dofs' must be boolean: "
                            "True or False"
                            % self.fixed_dofs)

    def _initialize_parameters(self, X, random_state):
        """Initialize the model parameters.

        Parameters
        ----------
        X : array-like, shape  (n_samples, n_features)

        random_state : RandomState
            A random number generator instance.
        """
        n_samples, _ = X.shape

        if self.init_params == 'kmeans':
            resp = np.zeros((n_samples, self.n_components))
            label = cluster.KMeans(n_clusters=self.n_components, n_init=1,
                                   random_state=random_state).fit(X).labels_
            resp[np.arange(n_samples), label] = 1
        elif self.init_params == 'random':
            resp = random_state.rand(n_samples, self.n_components)
            resp /= resp.sum(axis=1)[:, np.newaxis]
        else:
            raise ValueError("Unimplemented initialization method '%s'"
                             % self.init_params)
        gamma_priors = np.ones(resp.shape)

        self._initialize(X, resp, gamma_priors)

    def _initialize(self, X, resp, gamma_priors):
        """Initialization of the Student mixture parameters.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        resp : array-like, shape (n_samples, n_components)

        gamma_priors : array-like, shape (n_samples, n_components)
        """
        n_samples, n_features = X.shape

        weights, locations, scales = _estimate_student_parameters(
            X, resp, gamma_priors, self.reg_scale, self.scale_type)
        weights /= n_samples

        self.weights_ = (weights if self.weights_init is None
                         else self.weights_init)
        self.locations_ = (locations if self.locations_init is None
                           else self.locations_init)
        self.dofs_ = _initialize_dofs(
            X, scales, self.scale_type, self.n_components,
            resp, gamma_priors, n_features
        ) if self.dofs_init is None else self.dofs_init

        if self.precisions_init is None:
            self.scales_ = scales
            self.precisions_cholesky_ = _compute_precision_cholesky(
                scales, self.scale_type)
        elif self.scale_type == 'full':
            self.precisions_cholesky_ = np.array(
                [linalg.cholesky(prec_init, lower=True)
                 for prec_init in self.precisions_init])
        elif self.scale_type == 'tied':
            self.precisions_cholesky_ = linalg.cholesky(self.precisions_init,
                                                        lower=True)
        else:
            self.precisions_cholesky_ = self.precisions_init

    def fit_predict(self, X, y=None):
        """Estimate model parameters using X and predict the labels for X.

        The method fits the model n_init times and sets the parameters with
        which the model has the largest likelihood or lower bound. Within each
        trial, the method iterates between E-step and M-step for `max_iter`
        times until the change of likelihood or lower bound is less than
        `tol`, otherwise, a `ConvergenceWarning` is raised. After fitting, it
        predicts the most probable label for the input data points.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        labels : array, shape (n_samples,)
            Component labels.
        """
        X = _check_X(X, self.n_components, ensure_min_samples=2)
        self._check_initial_parameters(X)

        # if we enable warm_start, we will have a unique initialisation
        do_init = not(self.warm_start and hasattr(self, 'converged_'))
        n_init = self.n_init if do_init else 1

        max_lower_bound = -np.infty
        self.converged_ = False

        random_state = check_random_state(self.random_state)

        for init in range(n_init):
            self._print_verbose_msg_init_beg(init)

            if do_init:
                self._initialize_parameters(X, random_state)

            lower_bound = (-np.infty if do_init else self.lower_bound_)

            for n_iter in range(1, self.max_iter + 1):
                prev_lower_bound = lower_bound

                log_prob_norm, log_resp, gamma_priors = self._e_step(X)
                self._m_step(X, log_resp, gamma_priors)
                lower_bound = self._compute_lower_bound(
                    log_resp, log_prob_norm)

                change = lower_bound - prev_lower_bound
                self._print_verbose_msg_iter_end(n_iter, change)

                if abs(change) < self.tol:
                    self.converged_ = True
                    break

            self._print_verbose_msg_init_end(lower_bound)

            if lower_bound > max_lower_bound:
                max_lower_bound = lower_bound
                best_params = self._get_parameters()
                best_n_iter = n_iter

        if not self.converged_:
            warnings.warn('Initialization %d did not converge. '
                          'Try different init parameters, '
                          'or increase max_iter, tol '
                          'or check for degenerate data.'
                          % (init + 1), ConvergenceWarning)

        self._set_parameters(best_params)
        self.n_iter_ = best_n_iter
        self.lower_bound_ = max_lower_bound

        # Always do a final e-step to guarantee that the labels returned by
        # fit_predict(X) are always consistent with fit(X).predict(X)
        # for any value of max_iter and tol (and any random_state).
        _, log_resp, _ = self._e_step(X)

        return log_resp.argmax(axis=1)
        
    def _e_step(self, X):
        """E step.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        Returns
        -------
        log_prob_norm : float
            location of the logarithms of the probabilities of each sample in X

        log_responsibility : array, shape (n_samples, n_components)
            Logarithm of the posterior probabilities (or responsibilities) of
            the point of each sample in X.

        gamma_priors : array, shape (n_samples, n_components)
            Gamma weights of the point of each sample in X.
        """
        log_prob_norm, log_resp, gamma_priors = self._estimate_log_prob_resp(X)
        return np.mean(log_prob_norm), log_resp, gamma_priors

    def _m_step(self, X, log_resp, gamma_priors):
        """M step.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        log_resp : array-like, shape (n_samples, n_components)
            Logarithm of the posterior probabilities (or responsibilities) of
            the point of each sample in X.

        gamma_priors : array-like, shape (n_samples, n_components)
            Gamma weights of the point of each sample in X.
        """
        n_samples, n_features = X.shape
        resp = np.exp(log_resp)
        self.weights_, self.locations_, self.scales_ = (
            _estimate_student_parameters(X, resp, gamma_priors, self.reg_scale,
                                         self.scale_type))
        if not self.fixed_dofs:
            if self.algorithm == 'mcemc':
                _, log_resp, gamma_priors = self._e_step(X)
            self.dofs_ = _estimate_dofs(np.exp(log_resp), gamma_priors,
                                        self.dofs_, n_features)
        self.weights_ /= n_samples
        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.scales_, self.scale_type)

    def _estimate_log_prob(self, X):
        """Estimate the log-probability of the model."""
        return _estimate_log_student_prob(
            X, self.locations_, self.precisions_cholesky_,
            self.scale_type, self.dofs_)

    def _estimate_log_prob_resp(self, X):
        """Estimate log probabilities and responsibilities for each sample.

        Compute the log probabilities, weighted log probabilities per
        component and responsibilities for each sample in X with respect to
        the current state of the model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        Returns
        -------
        log_prob_norm : array, shape (n_samples,)
            log p(X)

        log_responsibilities : array, shape (n_samples, n_components)
            logarithm of the responsibilities

        gamma_priors : array, shape (n_samples, n_components)
            Gamma weights of the point of each sample in X.
        """
        weighted_log_prob = self._estimate_weighted_log_prob(X)
        log_prob_norm = logsumexp(weighted_log_prob, axis=1)
        with np.errstate(under='ignore'):
            # ignore underflow
            log_resp = weighted_log_prob - log_prob_norm[:, np.newaxis]
        gamma_priors = self._estimate_gamma_priors(X)
        return log_prob_norm, log_resp, gamma_priors

    def _estimate_gamma_priors(self, X):
        """Estimate the gamma priors.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        Returns
        -------
        gamma_priors : array, shape (n_samples, n_component)
        """
        return _compute_gamma_priors(X, self.locations_, self.precisions_cholesky_, self.scale_type, self.dofs_)

    def _estimate_log_weights(self):
        """Estimate the logarithm of the weights of the model."""
        return np.log(self.weights_)

    def _compute_lower_bound(self, _, log_prob_norm):
        """Returns the lower bound for the EM algorithm."""
        return log_prob_norm

    def _check_is_fitted(self):
        """Check that the model is fitted and the parameters have been set."""
        check_is_fitted(self, ['weights_', 'locations_', 'precisions_cholesky_', 'dofs_'])

    def _get_parameters(self):
        """Get the parameters of the model."""
        return (self.weights_, self.locations_, self.scales_,
                self.precisions_cholesky_, self.dofs_)

    def _set_parameters(self, params):
        """Set the parameters of the model."""
        (self.weights_, self.locations_, self.scales_,
         self.precisions_cholesky_, self.dofs_) = params

        # Attributes computation

        if self.scale_type == 'full':
            self.precisions_ = np.empty(self.precisions_cholesky_.shape)
            for k, prec_chol in enumerate(self.precisions_cholesky_):
                self.precisions_[k] = np.dot(prec_chol, prec_chol.T)

        elif self.scale_type == 'tied':
            self.precisions_ = np.dot(self.precisions_cholesky_,
                                      self.precisions_cholesky_.T)
        else:
            self.precisions_ = self.precisions_cholesky_ ** 2

    def _n_parameters(self):
        """Return the number of free parameters in the model."""
        _, n_features = self.locations_.shape
        if self.scale_type == 'full':
            scale_params = self.n_components * n_features * (n_features + 1) / 2.
        elif self.scale_type == 'diag':
            scale_params = self.n_components * n_features
        elif self.scale_type == 'tied':
            scale_params = n_features * (n_features + 1) / 2.
        else:
            scale_params = self.n_components
        location_params = n_features * self.n_components
        return int(scale_params + location_params + 2 * self.n_components - 1)

    def pdf(self, X):
        """Probability distribution function of the samples in X

        Parameters
        ----------
        X : array of shape (n_samples, n_features)

        Returns
        -------
        pdf : array of shape (n_samples,)
        """
        return np.exp(self.score_samples(X))

    def bic(self, X):
        """Bayesian information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)

        Returns
        -------
        bic : float
            The lower the better.
        """
        return (-2 * self.score(X) * X.shape[0] +
                self._n_parameters() * np.log(X.shape[0]))

    def aic(self, X):
        """Akaike information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)

        Returns
        -------
        aic : float
            The lower the better.
        """
        return -2 * self.score(X) * X.shape[0] + 2 * self._n_parameters()

    def predict_proba(self, X):
        """Predict posterior probability of each component given the data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        resp : array, shape (n_samples, n_components)
            Returns the probability of each mixture (state) in
            the model given each sample.
        """
        self._check_is_fitted()
        X = _check_X(X, None, self.locations_.shape[1])
        _, log_resp, _ = self._estimate_log_prob_resp(X)
        return np.exp(log_resp)

    def score_samples(self, X):
        """Compute the weighted log probabilities for each sample.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        log_prob : array, shape (n_samples,)
            Log probabilities of each data point in X.
        """
        self._check_is_fitted()
        X = _check_X(X, None, self.locations_.shape[1])

        return logsumexp(self._estimate_weighted_log_prob(X), axis=1)

    def predict(self, X):
        """Predict the labels for the data samples in X using trained model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        labels : array, shape (n_samples,)
            Component labels.
        """
        self._check_is_fitted()
        X = _check_X(X, None, self.locations_.shape[1])
        return self._estimate_weighted_log_prob(X).argmax(axis=1)

    def sample(self, n_samples=1):
        """Generate random samples from the fitted Student's t mixture distribution.

        Parameters
        ----------
        n_samples : int, optional
            Number of samples to generate. Defaults to 1.

        Returns
        -------
        X : array, shape (n_samples, n_features)
            Randomly generated sample

        y : array, shape (nsamples,)
            Component labels

        """
        self._check_is_fitted()

        if n_samples < 1:
            raise ValueError(
                "Invalid value for 'n_samples': %d . The sampling requires at "
                "least one sample." % self.n_components)

        _, n_features = self.locations_.shape
        rng = check_random_state(self.random_state)
        n_samples_comp = rng.multinomial(n_samples, self.weights_)

        if self.scale_type == 'full':
            X = np.vstack([
                _multivariate_t_random(location, scale, dof, int(sample), rng)
                for (location, scale, dof, sample) in zip(
                    self.locations_, self.scales_, self.dofs_, n_samples_comp)])
        elif self.scale_type == "tied":
            X = np.vstack([
                _multivariate_t_random(location, self.scales_, dof, int(sample), rng)
                for (location, dof, sample) in zip(
                    self.locations_, n_samples_comp)])
        elif self.scale_type == 'diag':
            X = np.vstack([
                _multivariate_t_random(location, np.diag(scale), dof, int(sample), rng)
                for (location, scale, dof, sample) in zip(
                    self.locations_, self.scales_, self.dofs_, n_samples_comp)])
        else:
            X = np.vstack([
                _multivariate_t_random(location, np.diag(self.scales_.repeat(n_features)), dof, int(sample), rng)
                for (location, dof, sample) in zip(
                    self.locations_, n_samples_comp)])

        y = np.concatenate([np.full(sample, j, dtype=int)
                           for j, sample in enumerate(n_samples_comp)])

        return X, y

    def rvs(self, n_samples=1):
        """Generate random samples from the fitted Student's t mixture distribution.

        Parameters
        ----------
        n_samples : int, optional
            Number of samples to generate. Defaults to 1.

        Returns
        -------
        X : array, shape (n_samples, n_features)
            Randomly generated sample

        """
        X, _ = self.sample(n_samples)

        return X

    def cdf(self, X, maxpts=1e+7, abseps=1e-6, releps=1e-6):
        """Cumulative distribution function of the samples in X

        Parameters
        ----------
        X : array of shape (n_samples, n_features)
        maxpts: integer
            The maximum number of points to use for integration (used when n_features > 3)
        abseps: float
            Absolute error tolerance (used when n_features > 1)
        releps: float
            Relative error tolerance (used when n_features == 2 or n_features == 3)

        Returns
        -------
        cdf : array of shape (n_samples,)
        """

        self._check_is_fitted()
        X = _check_X(X, None, self.locations_.shape[1])
        n_features = X.shape[1]
        f = np.zeros(X.shape[0])

        for i in range(self.n_components):
            location = self.locations_[i, :]
            dof = self.dofs_[i]
            weight = self.weights_[i]
            if self.scale_type == 'full':
                scale = self.scales_[i, :, :]
            elif self.scale_type == "tied":
                scale = self.scales_
            elif self.scale_type == 'diag':
                scale = np.diag(self.scales_[i, :])
            else:
                scale = np.diag(np.array([self.scales_[i]] * n_features))

            f = f + weight * _multivariate_t_cdf(X, location, scale, dof, maxpts=maxpts, abseps=abseps, releps=releps)

        f[f > 1.0] = 1.0
        f[f < 0.0] = 0.0
        return f



import numpy as np

from scipy import linalg
from scipy.special import gammaln, digamma, polygamma
from scipy.optimize import newton

from sklearn.utils.extmath import row_norms
from sklearn.utils import check_array


###############################################################################

# Student mixture shape checkers used by the StudentMixture class
def _check_shape(param, param_shape, name):
    """Validate the shape of the input parameter 'param'.

    Parameters
    ----------
    param : array

    param_shape : tuple

    name : string
    """
    param = np.array(param)
    if param.shape != param_shape:
        raise ValueError("The parameter '%s' should have the shape of %s, "
                         "but got %s" % (name, param_shape, param.shape))


def _check_X(X, n_components=None, n_features=None, ensure_min_samples=1):
    """Check the input data X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    n_components : int

    Returns
    -------
    X : array, shape (n_samples, n_features)
    """
    X = check_array(X, dtype=[np.float64, np.float32],
                    ensure_min_samples=ensure_min_samples)
    if n_components is not None and X.shape[0] < n_components:
        raise ValueError('Expected n_samples >= n_components '
                         'but got n_components = %d, n_samples = %d'
                         % (n_components, X.shape[0]))
    if n_features is not None and X.shape[1] != n_features:
        raise ValueError("Expected the input data X have %d features, "
                         "but got %d features"
                         % (n_features, X.shape[1]))
    return X


def _check_weights(weights, n_components):
    """Check the user provided 'weights'.

    Parameters
    ----------
    weights : array-like, shape (n_components,)
        The proportions of components of each mixture.

    n_components : int
        Number of components.

    Returns
    -------
    weights : array, shape (n_components,)
    """
    weights = check_array(weights, dtype=[np.float64, np.float32],
                          ensure_2d=False)
    _check_shape(weights, (n_components,), 'weights')

    # check range
    if (any(np.less(weights, 0.)) or
            any(np.greater(weights, 1.))):
        raise ValueError("The parameter 'weights' should be in the range "
                         "[0, 1], but got max value %.5f, min value %.5f"
                         % (np.min(weights), np.max(weights)))

    # check normalization
    if not np.allclose(np.abs(1. - np.sum(weights)), 0.):
        raise ValueError("The parameter 'weights' should be normalized, "
                         "but got sum(weights) = %.5f" % np.sum(weights))
    return weights


def _check_locations(locations, n_components, n_features):
    """Validate the provided 'locations'.

    Parameters
    ----------
    locations : array-like, shape (n_components, n_features)
        The centers of the current components.

    n_components : int
        Number of components.

    n_features : int
        Number of features.

    Returns
    -------
    locations : array, (n_components, n_features)
    """
    locations = check_array(locations, dtype=[np.float64, np.float32], ensure_2d=False)
    _check_shape(locations, (n_components, n_features), 'locations')
    return locations


def _check_dofs(dofs, n_components):
    """Check the user provided 'dofs'.

    Parameters
    ----------
    dofs : array-like, shape (n_components,)
        The degrees-of-freedom of each mixture.

    n_components : int
        Number of components.

    Returns
    -------
    weights : array, shape (n_components,)
    """
    dofs = check_array(dofs, dtype=[np.float64, np.float32], ensure_2d=False)
    _check_shape(dofs, (n_components,), 'dofs')

    # check range
    if any(np.less(dofs, 2)):
        raise ValueError("The parameter 'dofs' should be in the range "
                         "[2, inf), but got min value %.5f"
                         % (np.min(dofs)))
    return dofs


def _check_precision_positivity(precision, scale_type):
    """Check a precision vector is positive-definite."""
    if np.any(np.less_equal(precision, 0.0)):
        raise ValueError("'%s precision' should be "
                         "positive" % scale_type)


def _check_precision_matrix(precision, scale_type):
    """Check a precision matrix is symmetric and positive-definite."""
    if not (np.allclose(precision, precision.T) and
            np.all(linalg.eigvalsh(precision) > 0.)):
        raise ValueError("'%s precision' should be symmetric, "
                         "positive-definite" % scale_type)


def _check_precisions_full(precisions, scale_type):
    """Check the precision matrices are symmetric and positive-definite."""
    for prec in precisions:
        _check_precision_matrix(prec, scale_type)


def _check_precisions(precisions, scale_type, n_components, n_features):
    """Validate user provided precisions.

    Parameters
    ----------
    precisions : array-like
        'full' : shape of (n_components, n_features, n_features)
        'tied' : shape of (n_features, n_features)
        'diag' : shape of (n_components, n_features)
        'spherical' : shape of (n_components,)

    scale_type : string

    n_components : int
        Number of components.

    n_features : int
        Number of features.

    Returns
    -------
    precisions : array
    """
    precisions = check_array(precisions, dtype=[np.float64, np.float32],
                             ensure_2d=False,
                             allow_nd=scale_type == 'full')

    precisions_shape = {'full': (n_components, n_features, n_features),
                        'tied': (n_features, n_features),
                        'diag': (n_components, n_features),
                        'spherical': (n_components,)}
    _check_shape(precisions, precisions_shape[scale_type],
                 '%s precision' % scale_type)

    _check_precisions_type = {'full': _check_precisions_full,
                              'tied': _check_precision_matrix,
                              'diag': _check_precision_positivity,
                              'spherical': _check_precision_positivity}
    _check_precisions_type[scale_type](precisions, scale_type)
    return precisions


###############################################################################
# Student mixture parameters estimators (used by the M-Step)

def _estimate_student_scales_full(resp, gamma_priors, X, nk, locations, reg_scale):
    """Estimate the full scale matrices.

    Parameters
    ----------
    resp : array-like, shape (n_samples, n_components)
    
    gamma_priors : array-like, shape (n_samples, n_components)

    X : array-like, shape (n_samples, n_features)

    nk : array-like, shape (n_components,)

    locations : array-like, shape (n_components, n_features)

    reg_scale : float

    Returns
    -------
    scales : array, shape (n_components, n_features, n_features)
        The scale matrix of the current components.
    """
    n_components, n_features = locations.shape
    scales = np.empty((n_components, n_features, n_features))
    for k in range(n_components):
        diff = X - locations[k]
        scales[k] = np.dot(resp[:, k] * gamma_priors[:, k] * diff.T, diff) / nk[k]
        scales[k].flat[::n_features + 1] += reg_scale
    return scales


def _estimate_student_scales_tied(resp, gamma_priors, X, nk, locations, reg_scale):
    """Estimate the tied scale matrix.

    Parameters
    ----------
    resp : array-like, shape (n_samples, n_components)
    
    gamma_priors : array-like, shape (n_samples, n_components)

    X : array-like, shape (n_samples, n_features)

    nk : array-like, shape (n_components,)

    locations : array-like, shape (n_components, n_features)

    reg_scale : float

    Returns
    -------
    scale : array, shape (n_features, n_features)
        The tied scale matrix of the components.
    """
    avg_X2 = np.dot(X.T, X)
    avg_locations2 = np.dot(nk * locations.T, locations)
    scale = avg_X2 - avg_locations2
    scale /= nk.sum()
    scale.flat[::len(scale) + 1] += reg_scale
    return scale


def _estimate_student_scales_diag(resp, gamma_priors, X, nk, locations, reg_scale):
    """Estimate the diagonal scale vectors.

    Parameters
    ----------
    resp : array-like, shape (n_samples, n_components)
    
    gamma_priors : array-like, shape (n_samples, n_components)

    X : array-like, shape (n_samples, n_features)

    nk : array-like, shape (n_components,)

    locations : array-like, shape (n_components, n_features)

    reg_scale : float

    Returns
    -------
    scales : array, shape (n_components, n_features)
        The scale vector of the current components.
    """
    avg_X2 = np.dot((resp * gamma_priors).T, X * X) / nk[:, np.newaxis]
    avg_locations2 = locations ** 2
    avg_X_locations = locations * np.dot((resp * gamma_priors).T, X) / nk[:, np.newaxis]
    return avg_X2 - 2 * avg_X_locations + avg_locations2 + reg_scale


def _estimate_student_scales_spherical(resp, gamma_priors, X, nk, locations, reg_scale):
    """Estimate the spherical variance values.

    Parameters
    ----------
    resp : array-like, shape (n_samples, n_components)
    
    gamma_priors : array-like, shape (n_samples, n_components)

    X : array-like, shape (n_samples, n_features)

    nk : array-like, shape (n_components,)

    locations : array-like, shape (n_components, n_features)

    reg_scale : float

    Returns
    -------
    variances : array, shape (n_components,)
        The variance values of each components.
    """
    return _estimate_student_scales_diag(resp, gamma_priors, X, nk,
                                         locations, reg_scale).mean(1)


def _estimate_student_parameters(X, resp, gamma_priors, reg_scale, scale_type):
    """Estimate the Student's t-distribution parameters.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        The input data array.

    resp : array-like, shape (n_samples, n_components)
        The responsibilities for each data sample in X.

    gamma_priors : array-like, shape (n_samples, n_components)
        The gamma priors for each data sample in X.

    reg_scale : float
        The regularization added to the diagonal of the scale matrices.

    scale_type : {'full', 'tied', 'diag', 'spherical'}
        The type of precision matrices.

    Returns
    -------
    nk : array-like, shape (n_components,)
        The numbers of data samples in the current components.

    locations : array-like, shape (n_components, n_features)
        The centers of the current components.

    scales : array-like
        The scale matrix of the current components.
        The shape depends on scale_type.
    """
    nk = resp.sum(axis=0) + 10 * np.finfo(resp.dtype).eps
    locations = np.dot((resp * gamma_priors).T, X) / (resp * gamma_priors).T.sum(1)[:, np.newaxis]
    scales = {"full": _estimate_student_scales_full,
              "tied": _estimate_student_scales_tied,
              "diag": _estimate_student_scales_diag,
              "spherical": _estimate_student_scales_spherical
              }[scale_type](resp, gamma_priors, X, nk, locations, reg_scale)
    return nk, locations, scales


def _compute_precision_cholesky(scales, scale_type):
    """Compute the Cholesky decomposition of the precisions.

    Parameters
    ----------
    scales : array-like
        The scale matrix of the current components.
        The shape depends of the scale_type.

    scale_type : {'full', 'tied', 'diag', 'spherical'}
        The type of precision matrices.

    Returns
    -------
    precisions_cholesky : array-like
        The cholesky decomposition of sample precisions of the current
        components. The shape depends of the scale_type.
    """
    estimate_precision_error_message = (
        "Fitting the mixture model failed because some components have "
        "ill-defined empirical scale (for instance caused by singleton "
        "or collapsed samples). Try to decrease the number of components, "
        "or increase reg_scale.")

    if scale_type in 'full':
        n_components, n_features, _ = scales.shape
        precisions_chol = np.empty((n_components, n_features, n_features))
        for k, scale in enumerate(scales):
            try:
                cov_chol = linalg.cholesky(scale, lower=True)
            except linalg.LinAlgError:
                raise ValueError(estimate_precision_error_message)
            precisions_chol[k] = linalg.solve_triangular(cov_chol,
                                                         np.eye(n_features),
                                                         lower=True).T
    elif scale_type == 'tied':
        _, n_features = scales.shape
        try:
            cov_chol = linalg.cholesky(scales, lower=True)
        except linalg.LinAlgError:
            raise ValueError(estimate_precision_error_message)
        precisions_chol = linalg.solve_triangular(cov_chol, np.eye(n_features),
                                                  lower=True).T
    else:
        if np.any(np.less_equal(scales, 0.0)):
            print(scales)
            raise ValueError(estimate_precision_error_message)
        precisions_chol = 1. / np.sqrt(scales)
    return precisions_chol


def _compute_gamma_priors(X, locations, precisions_chol, scale_type, dofs):
    """Estimate the gamma weights of samples.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    locations : array-like, shape (n_components, n_features)

    precisions_chol : array-like
        Cholesky decompositions of the precision matrices.
        'full' : shape of (n_components, n_features, n_features)
        'tied' : shape of (n_features, n_features)
        'diag' : shape of (n_components, n_features)
        'spherical' : shape of (n_components,)

    scale_type : {'full', 'tied', 'diag', 'spherical'}

    dofs : array-like, shape (n_components,)

    Returns
    -------
    gamma_weights : array, shape (n_samples, n_components)
    """
    n_samples, n_features = X.shape
    n_components, _ = locations.shape

    if scale_type == 'full':
        prob = np.empty((n_samples, n_components))
        for k, (mu, prec_chol) in enumerate(zip(locations, precisions_chol)):
            y = np.dot(X, prec_chol) - np.dot(mu, prec_chol)
            prob[:, k] = np.sum(np.square(y), axis=1)

    elif scale_type == 'tied':
        prob = np.empty((n_samples, n_components))
        for k, mu in enumerate(locations):
            y = np.dot(X, precisions_chol) - np.dot(mu, precisions_chol)
            prob[:, k] = np.sum(np.square(y), axis=1)

    elif scale_type == 'diag':
        precisions = precisions_chol ** 2
        prob = (np.sum((locations ** 2 * precisions), 1) -
                2. * np.dot(X, (locations * precisions).T) +
                np.dot(X ** 2, precisions.T))

    else:
        precisions = precisions_chol ** 2
        prob = (np.sum(locations ** 2, 1) * precisions -
                2 * np.dot(X, locations.T * precisions) +
                np.outer(row_norms(X, squared=True), precisions))

    return (dofs + n_features) / (dofs + prob)


###############################################################################
# Student mixture probability estimators
def _compute_log_det_cholesky(matrix_chol, scale_type, n_features):
    """Compute the log-det of the cholesky decomposition of matrices.

    Parameters
    ----------
    matrix_chol : array-like
        Cholesky decompositions of the matrices.
        'full' : shape of (n_components, n_features, n_features)
        'tied' : shape of (n_features, n_features)
        'diag' : shape of (n_components, n_features)
        'spherical' : shape of (n_components,)

    scale_type : {'full', 'tied', 'diag', 'spherical'}

    n_features : int
        Number of features.

    Returns
    -------
    log_det_precision_chol : array-like, shape (n_components,)
        The determinant of the precision matrix for each component.
    """
    if scale_type == 'full':
        n_components, _, _ = matrix_chol.shape
        log_det_chol = (np.sum(np.log(
            matrix_chol.reshape(
                n_components, -1)[:, ::n_features + 1]), 1))

    elif scale_type == 'tied':
        log_det_chol = (np.sum(np.log(np.diag(matrix_chol))))

    elif scale_type == 'diag':
        log_det_chol = (np.sum(np.log(matrix_chol), axis=1))

    else:
        log_det_chol = n_features * (np.log(matrix_chol))

    return log_det_chol


def _estimate_log_student_prob(X, locations, precisions_chol, scale_type, dofs):
    """Estimate the log Student probability.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    locations : array-like, shape (n_components, n_features)

    precisions_chol : array-like
        Cholesky decompositions of the precision matrices.
        'full' : shape of (n_components, n_features, n_features)
        'tied' : shape of (n_features, n_features)
        'diag' : shape of (n_components, n_features)
        'spherical' : shape of (n_components,)

    scale_type : {'full', 'tied', 'diag', 'spherical'}

    dofs : array-like, shape (n_components,)

    Returns
    -------
    log_prob : array, shape (n_samples, n_components)
    """
    n_samples, n_features = X.shape
    n_components, _ = locations.shape
    # det(precision_chol) is half of det(precision)
    log_det = _compute_log_det_cholesky(
        precisions_chol, scale_type, n_features)

    if scale_type == 'full':
        prob = np.empty((n_samples, n_components))
        for k, (mu, prec_chol) in enumerate(zip(locations, precisions_chol)):
            y = np.dot(X, prec_chol) - np.dot(mu, prec_chol)
            prob[:, k] = np.sum(np.square(y), axis=1)

    elif scale_type == 'tied':
        prob = np.empty((n_samples, n_components))
        for k, mu in enumerate(locations):
            y = np.dot(X, precisions_chol) - np.dot(mu, precisions_chol)
            prob[:, k] = np.sum(np.square(y), axis=1)

    elif scale_type == 'diag':
        precisions = precisions_chol ** 2
        prob = (np.sum((locations ** 2 * precisions), 1) -
                2. * np.dot(X, (locations * precisions).T) +
                np.dot(X ** 2, precisions.T))

    else:
        precisions = precisions_chol ** 2
        prob = (np.sum(locations ** 2, 1) * precisions -
                2 * np.dot(X, locations.T * precisions) +
                np.outer(row_norms(X, squared=True), precisions))
    log_prob = np.log(1 + (prob / dofs))
    return (gammaln(0.5*(dofs + n_features)) - gammaln(0.5 * dofs) + log_det -
            0.5 * (n_features * np.log(dofs * np.pi) + (dofs + n_features) * log_prob))


def _initialize_dofs(X, scales, scale_type, n_components, resp, gamma_priors, n_features, max_iter=100, tol=1e-3):
    """Initializes the degrees-of-freedom.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    scales : array-like
        scale matrices.
        'full' : shape of (n_components, n_features, n_features)
        'tied' : shape of (n_features, n_features)
        'diag' : shape of (n_components, n_features)
        'spherical' : shape of (n_components,)

    scale_type : {'full', 'tied', 'diag', 'spherical'}

    n_components : int
        Number of components.

    resp : array-like, shape (n_samples, n_components)
        The responsibilities for each data sample in X.

    gamma_priors : array-like, shape (n_samples, n_components)
        The gamma priors for each data sample in X.

    n_features : int
        Number of features.

    max_iter : int, defaults to 100.
        The number of iterations to perform for degrees-of-freedom estimation.

    tol : float, defaults to 1e-3.
        The degrees-of-freedom estimation convergence threshold.

    Returns
    -------
    dofs : array-like, shape (n_components,)
        Initial degrees-of-freedom estimation.
    """
    arg_resp = resp.argmax(1)
    current_dofs = np.zeros((n_components,))
    for k in range(n_components):
        covariance = np.cov(X[arg_resp == k, :].T)
        if scale_type == 'full':
            scale = np.diag(scales[k, :, :])
        elif scale_type == 'tied':
            scale = np.diag(scales)
        elif scale_type == 'diag':
            scale = scales[k, :]
        else:
            scale = np.array(n_features * [scales[k]])
        dof_estimation = 2 / (1 - np.mean(scale * np.diag(np.linalg.pinv(covariance))))
        current_dofs[k] = np.minimum(np.maximum(2, dof_estimation), 30)
    return _estimate_dofs(resp, gamma_priors, current_dofs, n_features, max_iter=max_iter, tol=tol)


def _estimate_dofs(resp, gamma_priors, current_dofs, n_features, max_iter=100, tol=1e-3):
    """Estimates the degrees-of-freedom.

    Parameters
    ----------
    resp : array-like, shape (n_samples, n_components)
        The responsibilities for each data sample in X.

    gamma_priors : array-like, shape (n_samples, n_components)
        The gamma priors for each data sample in X.

    current_dofs : array-like, shape (n_components,)
        Current degrees-of-freedom estimation.

    n_features : int
        Number of features.

    max_iter : int, defaults to 100.
        The number of iterations to perform for degrees-of-freedom estimation.

    tol : float, defaults to 1e-3.
        The degrees-of-freedom estimation convergence threshold.

    Returns
    -------
    dofs : array-like, shape (n_components,)
        The degrees-of-freedom estimation.
    """
    n_components = current_dofs.shape[0]
    dofs = np.zeros((n_components,))
    for k in range(n_components):
        constant = (
                1.0
                + (1.0 / resp[:, k].sum(axis=0))
                * (resp[:, k] * (np.log(gamma_priors[:, k]) - gamma_priors[:, k])).sum(axis=0)
                + digamma(1.0 * (current_dofs[k] + n_features) / 2.0)
                - np.log(1.0 * (current_dofs[k] + n_features) / 2.0)
        )

        def function(df):
            return -digamma(1.0 * df / 2.0) + np.log(1.0 * df / 2.0) + constant

        def first_derivative(df):
            return -(1.0 / 2.0) * polygamma(1, 1.0 * df / 2.0) + 1.0 / df

        def second_derivative(df):
            return -(1.0 / 4.0) * polygamma(2, 1.0 * df / 2.0) - 1.0 / (df * df)

        dofs[k] = newton(
            function,
            current_dofs[k],
            first_derivative,
            args=(),
            maxiter=max_iter,
            tol=tol,
            fprime2=second_derivative,
            full_output=False,
            disp=False,
        )

    return dofs



# -*- coding: utf-8 -*-

"""Top-level package for Student's t-mixture model."""

__author__ = """Omri Tomer"""
__email__ = 'omritomer1@mail.tau.ac.il'
__version__ = '0.1.4'

import sys

if (sys.version_info < (3, 3)):
    from student_mixture import StudentMixture
    from multivariate_t import multivariate_t as MultivariateT
    from mutlivariate_t_fit import MultivariateTFit

else:
    from student_mixture.student_mixture import StudentMixture
    from student_mixture.multivariate_t import multivariate_t  as MultivariateT
    from student_mixture.mutlivariate_t_fit import MultivariateTFit

__all__ = ['StudentMixture',
           'MultivariateT',
           'MultivariateTFit']



import numpy as np
from scipy.linalg import cholesky
from scipy.integrate import quad
from scipy.special import betainc
from .multivariate_t_mc import _multivariate_t_cdf_single_sample

###############################################################################
# Functions that estimate the cumulative distribution function of the standard multivariate t-distribution


def _standard_t_cdf_univariate(x, dof, abseps=1e-8, releps=1e-8):
    """
    Standard univariate Student's t cumulative density function.
    Reference:
    Johnson, N. L., Kotz, S., Balakrishnan, N. ( 1995 ). Continuous
    Univariate Distributions. Vol. 2, 2nd ed. New York : Wiley.

    Parameters
    ----------
    x : array_like
        Standardized samples, shape (n_samples, )
    dof : float
        Degrees-of-freedom of the distribution, must be a positive real number
    abseps: float
        Absolute error tolerance for integration
    releps: float
        Relative error tolerance for integration


    Returns
    -------
    cdf : ndarray or scalar
        Cumulative density function evaluated at `x`
    """
    return 1.0 - 0.5 * betainc(dof / 2.0, 0.5, dof / (dof + x ** 2))


def _standard_t_cdf_bivariate(x, corr_mat, dof, abseps=1e-8, releps=1e-8):
    """
    Standard bivariate Student's t cumulative density function.
    Algorithm based on:
    Genz, A. (2004). Numerical computation of rectangular bivariate
    and trivariate normal and t probabilities. Statistics and
    Computing, 14(3), 251-260.

    Parameters
    ----------
    x : array_like
        Standardized samples, shape (n_samples, 2)
    corr_mat : array_like, shape (2, 2)
        Correlation matrix of the distribution,must be symmetric and positive
        definite, with all elements of the diagonal being 1
    dof : float
        Degrees-of-freedom of the distribution, must be a positive real number
    abseps: float
        Absolute error tolerance for integration
    releps: float
        Relative error tolerance for integration


    Returns
    -------
    cdf : ndarray or scalar
        Cumulative density function evaluated at `x`
    """

    rho = corr_mat[0, 1]
    n = x.shape[0]
    if rho >= 0:
        tau_s = _standard_t_cdf_univariate(x.min(axis=1), dof)
    else:
        tau_s = np.maximum(_standard_t_cdf_univariate(x[:, 0], dof) - _standard_t_cdf_univariate(-x[:, 1], dof), 0)

    integral = np.zeros_like(tau_s)
    lower_bound = np.sign(rho) * np.pi / 2.0 if rho != 0 else np.pi / 2.0
    upper_bound = np.arcsin(rho)

    for i in range(n):
        x1 = x[i, 0]
        x2 = x[i, 1]
        if np.isfinite(x1) and np.isfinite(x2):
            integral[i] = quad(bivariate_integrand, lower_bound, upper_bound, (x1, x2, dof),
                               epsabs=abseps, epsrel=releps)[0]
    return tau_s + integral / (2 * np.pi)


def _standard_t_cdf_multivariate(x, corr_mat, dof, tol=1e-4, max_evaluations=1e+7):
    """
    Wrapper function for .
    See multivariate_t_mc for more on algorithm

    Parameters
    ----------
    x : array_like
        Standardized samples, shape (n_samples, n_features)
    corr_mat : array_like, shape (n_features, n_features)
        Correlation matrix of the distribution,must be symmetric and positive
        definite, with all elements of the diagonal being 1
    dof : float
        Degrees-of-freedom of the distribution, must be a positive real number
    tol: float
        Tolerance for quasi-Monte Carlo algorithm
    max_evaluations: float
        Maximum points to evaluate with quasi-Monte Carlo algorithm
    

    Returns
    -------
    cdf : ndarray or scalar
        Cumulative density function evaluated at `x`
    """
    n = x.shape[0]
    cdf = np.zeros(n)
    err = np.zeros(n)
    chol = cholesky(corr_mat)
    for i in range(n):
        cdf[i], err[i] = _multivariate_t_cdf_single_sample(x[i, :], chol, dof, tol=tol, max_evaluations=max_evaluations)
    return cdf

###############################################################################
# Integral functions


def bivariate_integrand(theta, b1, b2, nu):
    sin_theta = np.sin(theta)
    cos_theta_squared = np.cos(theta) ** 2
    return (1 + ((b1 ** 2) + (b2 ** 2) - 2 * sin_theta * b1 * b2)
            / (nu * cos_theta_squared)) ** (-nu / 2.0)



"""Student's t-distribution Fitting."""

# Author: Omri Tomer <omritomer1@mail.tau.ac.il>
# License: BSD 3 clause

import numpy as np

from scipy import linalg

from sklearn.utils.validation import check_is_fitted

import warnings

from sklearn.exceptions import ConvergenceWarning
from sklearn.utils import check_random_state
from sklearn.utils.fixes import logsumexp

from .multivariate_t_fit_functions import _check_X, _check_location, _check_dof, _check_precision,\
    _estimate_student_parameters, _compute_precision_cholesky, _compute_gamma_priors, _estimate_log_student_prob,\
    _initialize_dof, _estimate_dof
from .multivariate_t import multivariate_t


class MultivariateTFit():
    """Multivariate Studnet's t-distribution fitting object.

    Class to fit a multivariate Student's t-distribution.
    This class allows estimation of the parameters of a
    multivariate Student's t-distribution.

    Parameters
    ----------
    scale_type : {'full' (default), 'diag', 'spherical'}
        String describing the type of scale parameters to use.
        Must be one of:

        'full'
            full scale matrix
        'diag'
            diagonal scale matrix
        'spherical'
            single variance

    algorithm : {'em' (default), 'mcecm'}
        String describing the algorithm used for estimating
        the degrees-of-freedom.
        Must be one of:

        'em'
            Expectation-Maximization algorithm
        'mcecm'
            Multicycle Expectation-Conditional-Maximization algorithm

    fixed_dof : boolean, default to False.
        Determines whether the degrees-of-freedom are estimated
        or fixed. If fixed, then the default value for 'dof_init'
        is np.inf

    tol : float, defaults to 1e-3.
        The convergence threshold. EM iterations will stop when the
        lower bound average gain is below this threshold.

    reg_scale : float, defaults to 1e-6.
        Non-negative regularization added to the diagonal of scale.
        Allows to assure that the scale matrix is positive.

    max_iter : int, defaults to 100.
        The number of EM iterations to perform.

    init_params : {'gaussian', 'random'}, defaults to 'gaussian'.
        The method used to initialize the weights, the means and the
        precisions.
        Must be one of::

            'gaussian' : responsibilities are initialized using Gaussian distribution.
            'random' : responsibilities are initialized randomly.

    dof_tol : float, defaults to 1e-3.
        The degrees-of-freedom estimation convergence threshold.

    dof_max_iter : int, defaults to 100.
        The number of iterations to perform for degrees-of-freedom estimation.

    location : array-like, shape (n_features,), optional
        The user-provided initial location, defaults to None,
        If it None, location is initialized using the `init_params` method.

    precision_init : array-like, optional.
        The user-provided initial precisions (inverse of the scale
        matrix), defaults to None.
        If it None, precision is initialized using the 'init_params' method.
        The shape depends on 'scale_type'::

            (1,)                        if 'spherical',
            (n_features,)               if 'diag',
            (n_features, n_features)    if 'full'

    dof_init : scalar (2 <= dof < inf), optional
        The user-provided initial degrees-of freedom, defaults to None.
        If None, dof is initialized using the `init_params` method.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    verbose : int, default to 0.
        Enable verbose output. If 1 then it prints the current
        initialization and each iteration step. If greater than 1 then
        it prints also the log probability and the time needed
        for each step.

    verbose_interval : int, default to 10.
        Number of iteration done before the next print.

    output : boolean, default to False.
        Determines the method 'fit' returns a multivariate_t
        object with the estimated parameters

    Attributes
    ----------
    location_ : array-like, shape (n_features,)
        The location (mean) of the t-distribution.

    scale_ : array-like
        The scale of the t-distribution.
        The shape depends on `scale_type`::

            (1,)                        if 'spherical',
            (n_features,)               if 'diag',
            (n_features, n_features)    if 'full'

    dof_ : scalar (2 <= dof_ < inf)
        The degrees-of-freedom of the t-distribution.

    precision_ : array-like
        The precision matrix of the t-distribution. A precision
        matrix is the inverse of the scale matrix. The scale matrix is
        symmetric positive definite so the Student's t-distribution
        can be equivalently parameterized by the precision matrix.
        Storing the precision matrix instead of the scale matrix makes
        it more efficient to compute the log-likelihood of new samples at test
        time. The shape depends on `scale_type`::

            (1,)                        if 'spherical',
            (n_features,)               if 'diag',
            (n_features, n_features)    if 'full'

    precisions_cholesky_ : array-like
        The cholesky decomposition of the precision matrix of the
        distribution. A precision matrix is the inverse of a scale matrix.
        A scale matrix is symmetric positive definite the Student's
        t-distributions can be equivalently parameterized by the
        precision matriX.
        Storing the precision matrix instead of the scale matrix makes
        it more efficient to compute the log-likelihood of new samples at test
        time. The shape depends on `scale_type`::

            (1,)                        if 'spherical',
            (n_features,)               if 'diag',
            (n_features, n_features)    if 'full'

    converged_ : bool
        True when convergence was reached in fit(), False otherwise.

    n_iter_ : int
        Number of step used by the best fit of EM to reach the convergence.

    lower_bound_ : float
        Lower bound value on the log-likelihood (of the training data with
        respect to the model) of the best fit of EM.


    """

    def __init__(self, scale_type='full', algorithm='em',
                 fixed_dof=False,
                 tol=1e-3, reg_scale=1e-6, max_iter=100,
                 dof_tol=1e-3, dof_max_iter=100,
                 location_init=None, precision_init=None, dof_init=None,
                 random_state=None,
                 verbose=0, verbose_interval=10,
                 output=False):

        self.scale_type = scale_type
        self.algorithm = algorithm
        self.fixed_dof = fixed_dof
        self.dof_tol = dof_tol
        self.dof_max_iter = dof_max_iter
        self.location_init = location_init
        self.precision_init = precision_init
        self.dof_init = dof_init
        self.tol = tol
        self.reg_scale = reg_scale
        self.max_iter = max_iter
        self.random_state = random_state
        self.verbose = verbose
        self.verbose_interval = verbose_interval
        self.output = output

    def _check_parameters(self, X):
        """Check that the parameters are well defined."""
        n_features = X.shape[1]
        if self.scale_type not in ['spherical', 'diag', 'full']:
            raise ValueError("Invalid value for 'scale_type': %s "
                             "'scale_type' should be in "
                             "['spherical', 'diag', 'full']"
                             % self.scale_type)

        if self.location_init is not None:
            self.location_init = _check_location(self.location_init, n_features)

        if self.precision_init is not None:
            self.precision_init = _check_precision(self.precision_init,
                                                   self.scale_type,
                                                   n_features)
        if self.dof_init is not None:
            self.dof_init = _check_dof(self.dof_init)

    def _check_initial_parameters(self, X):
        """Check values of the basic parameters.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        """

        if self.algorithm not in ['em', 'mcecm']:
            raise ValueError("Invalid value for 'algorithm': %s "
                             "'algorithm' should be in "
                             "['em', 'mcecm']"
                             % self.algorithm)

        if not isinstance(self.fixed_dof, bool):
            raise TypeError("Invalid value for 'fixed_dof': %s "
                            "'fixed_dof' must be boolean: "
                            "True or False"
                            % self.fixed_dof)
        if self.tol < 0.:
            raise ValueError("Invalid value for 'tol': %.5f "
                             "Tolerance used by the EM must be non-negative"
                             % self.tol)

        if self.max_iter < 1:
            raise ValueError("Invalid value for 'max_iter': %d "
                             "Estimation requires at least one iteration"
                             % self.max_iter)

        if self.reg_scale < 0.:
            raise ValueError("Invalid value for 'reg_scale': %.5f "
                             "regularization on scale must be "
                             "non-negative"
                             % self.reg_scale)

        if not isinstance(self.output, bool):
            raise TypeError("Invalid value for 'output': %s "
                            "'output' must be boolean: "
                            "True or False"
                            % self.output)

        # Check all the parameters values of the derived class
        self._check_parameters(X)

    def fit(self, X, y=None):
        """Estimate model parameters with the EM algorithm.

        The method fits the model. The method iterates between E-step and
        M-step for ``max_iter`` times until the change of likelihood or
        lower bound is less than ``tol``, otherwise, a ``ConvergenceWarning``
        is raised.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        self
        """
        self.fit_predict(X, y)

    def fit_predict(self, X, y=None):
        """Estimate model parameters using X and predict the labels for X.

        The method fits the model. The method iterates between E-step and
        M-step for ``max_iter`` times until the change of likelihood or
        lower bound is less than ``tol``, otherwise, a ``ConvergenceWarning``
        is raised.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        """
        X = _check_X(X, ensure_min_samples=2)
        self._check_initial_parameters(X)

        self.converged_ = False

        self._initialize(X)

        lower_bound = -np.infty

        for n_iter in range(1, self.max_iter + 1):
            prev_lower_bound = lower_bound

            log_prob_norm, gamma_priors = self._e_step(X)
            self._m_step(X, gamma_priors)
            lower_bound = self._compute_lower_bound(log_prob_norm)

            change = lower_bound - prev_lower_bound

            if abs(change) < self.tol:
                self.converged_ = True
                break

        if not self.converged_:
            warnings.warn('Initialization did not converge. '
                          'Try different init parameters, '
                          'or increase max_iter, tol '
                          'or check for degenerate data.',
                          ConvergenceWarning)

        self._set_parameters(self._get_parameters())
        self.n_iter_ = n_iter
        self.lower_bound_ = lower_bound
        if self.output:
            return self._dist

    def _initialize(self, X):
        """Initialization of the Student's t-distribution parameters.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        """
        n_samples, _ = X.shape
        location, scale = _estimate_student_parameters(
            X, np.ones(n_samples), self.reg_scale, self.scale_type)

        self.location_ = location if self.location_init is None else self.location_init
        if self.precision_init is None:
            self.scale_ = scale
            self.precision_cholesky_ = _compute_precision_cholesky(
                scale, self.scale_type)
        elif self.scale_type == 'full':
            self.precision_cholesky_ = linalg.cholesky(self.precision_init, lower=True)
        elif self.scale_type == 'tied':
            self.precision_cholesky_ = linalg.cholesky(self.precision_init, lower=True)
        else:
            self.precision_cholesky_ = self.precision_init

        self.dof_ = _initialize_dof(X, self.location_, self.precision_cholesky_,self.scale_type
                                    ) if self.dof_init is None else self.dof_init

    def _e_step(self, X):
        """E step.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        Returns
        -------
        log_prob_norm : float
            Mean of the logarithms of the probabilities of each sample in X

        gamma_priors : array, shape (n_samples,)
            Gamma weights of each sample in X.
        """
        log_prob_norm, gamma_priors = self._estimate_log_prob_gamma(X)
        self.gamma_priors_ = gamma_priors
        return log_prob_norm, gamma_priors

    def _estimate_log_prob_gamma(self, X):
        """Estimate log probabilities and responsibilities for each sample.

        Compute the log probability, and the prior Gamma weightsof the samples
        in X with respect to the current state of the model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        Returns
        -------
        log_prob_norm : float
            Mean of the logarithms of the probabilities of each sample in X

        gamma_priors : array, shape (n_samples,)
            Gamma weights of each sample in X.
        """
        log_prob = self._estimate_log_prob(X)
        log_prob_norm = logsumexp(log_prob)
        gamma_priors = self._estimate_gamma_priors(X)
        return log_prob_norm, gamma_priors

    def _estimate_gamma_priors(self, X):
        """Estimate the gamma priors.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        Returns
        -------
        gamma_priors : array, shape (n_samples,)
            Gamma weights of each sample in X.
        """
        return _compute_gamma_priors(X, self.location_, self.precision_cholesky_, self.scale_type, self.dof_)

    def _m_step(self, X, gamma_priors):
        """M step.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        gamma_priors : array, shape (n_samples,)
            Gamma weights of each sample in X.
        """
        _, n_features = X.shape
        self.location_, self.scale_ = (
            _estimate_student_parameters(X, gamma_priors, self.reg_scale,
                                          self.scale_type))

        if not self.fixed_dof:
            if self.algorithm == 'mcecm':
                _, gamma_priors = self._e_step(X)
            self.dof_ = _estimate_dof(gamma_priors, self.dof_, n_features)

        self.precision_cholesky_ = _compute_precision_cholesky(
            self.scale_, self.scale_type)

    def _estimate_log_prob(self, X):
        """Estimate the log-probability of the model."""
        return _estimate_log_student_prob(
            X, self.location_, self.precision_cholesky_, self.scale_type, self.dof_)

    def _compute_lower_bound(self, log_prob_norm):
        """Returns the lower bound for the EM algorithm."""
        return log_prob_norm

    def _check_is_fitted(self):
        """Check that the model is fitted and the parameters have been set."""
        check_is_fitted(self, ['dof_', 'location_', 'precision_cholesky_'])

    def _get_parameters(self):
        """Get the parameters of the model."""
        return (self.location_, self.scale_,
                self.precision_cholesky_, self.dof_)

    def _set_parameters(self, params):
        """Set the parameters of the model."""
        (self.location_, self.scale_,
         self.precision_cholesky_, self.dof_) = params
        self._dist = multivariate_t(self.location_, self.scale_, self.dof_)

        if self.scale_type == 'full':
            self.precision_ = np.dot(self.precision_cholesky_, self.precision_cholesky_.T)
        else:
            self.precision_ = self.precision_cholesky_ ** 2

    def _n_parameters(self):
        """Return the number of free parameters in the model."""
        n_features = self.location_.shape[0]
        if self.scale_type == 'full':
            cov_params = n_features * (n_features + 1) / 2.
        elif self.scale_type == 'diag':
            cov_params = n_features
        else:
            cov_params = 1
        return int(cov_params + n_features + 1)

    def pdf(self, X):
        """Bayesian information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_dimensions)

        Returns
        -------
        bic : float
            The lower the better.
        """
        return self._dist.pdf(X)

    def bic(self, X):
        """Bayesian information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_dimensions)

        Returns
        -------
        bic : float
            The lower the better.
        """
        return (-2 * self.log_likelihood(X) +
                self._n_parameters() * np.log(X.shape[0]))

    def aic(self, X):
        """Akaike information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_dimensions)

        Returns
        -------
        aic : float
            The lower the better.
        """
        return -2 * self.log_likelihood(X) + 2 * self._n_parameters()

    def log_likelihood(self, X):
        """Compute the per-sample average log-likelihood of the given data X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_dimensions)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        log_likelihood : float
            Log likelihood of the distribution given X.
        """
        return self._dist.logpdf(X).sum()

    def rvs(self, n_samples=1):
        """Generate random samples from the fitted Multivariate t-distribution.

        Parameters
        ----------
        n_samples : int, optional
            Number of samples to generate. Defaults to 1.

        Returns
        -------
        X : array, shape (n_samples, n_features)
            Randomly generated sample

        """
        self._check_is_fitted()

        if n_samples < 1:
            raise ValueError(
                "Invalid value for 'n_samples': %d . The sampling requires at "
                "least one sample." % (n_samples))

        rng = check_random_state(self.random_state)

        return self._dist.rvs(size=n_samples, random_state=rng)



#
# Author: Omri Tomer 2020
#
from __future__ import division, print_function, absolute_import

import numpy as np

from scipy._lib import doccer
from scipy.special import gammaln, digamma

from scipy.stats._multivariate import (multi_rv_generic, multi_rv_frozen, _doc_random_state,
                                       multivariate_normal, _PSD, _squeeze_output)
from .multivariate_t_functions import _multivariate_t_random, _multivariate_t_cdf

_LOG_2PI = np.log(2 * np.pi)
_LOG_PI = np.log(np.pi)

_mvt_doc_default_callparams = """\
location : array_like, optional
    Location of the distribution (default zero)
scale : array_like, optional
    Scale matrix of the distribution (default one)
dof : scalar, dof > 2, optional
    Degrees-of-freedom of the distribution (default np.inf)
allow_singular : bool, optional
    Whether to allow a singular covariance matrix.  (Default: False)
"""

_mvt_doc_callparams_note = \
    """Setting the parameter `location` to `None` is equivalent to having `location`
    be the zero-vector. The parameter `scale` can be a scalar, in which case
    the scale matrix is the identity times that value, a vector of
    diagonal entries for the scale matrix, or a two-dimensional
    array_like. The parameter 'dof' can be a scalar larger than 2. Setting it to
    'None' is equivalent to np.inf, in which case the distribution is equivalent to
    a multivariate normal distribution with mean=location and covariance=scale.
    """

_mvt_doc_frozen_callparams = ""

_mvt_doc_frozen_callparams_note = \
    """See class definition for a detailed description of parameters."""

mvt_docdict_params = {
    '_mvt_doc_default_callparams': _mvt_doc_default_callparams,
    '_mvt_doc_callparams_note': _mvt_doc_callparams_note,
    '_doc_random_state': _doc_random_state
}

mvt_docdict_noparams = {
    '_mvt_doc_default_callparams': _mvt_doc_frozen_callparams,
    '_mvt_doc_callparams_note': _mvt_doc_frozen_callparams_note,
    '_doc_random_state': _doc_random_state
}


class multivariate_t_gen(multi_rv_generic):
    r"""
    A multivariate Student's t-distribution random variable.

    The `location` keyword specifies the location. The `scale` keyword specifies the
    scale matrix. The 'dof' keyword specifies the degrees-of-freedom.

    Methods
    -------
    ``pdf(x, location=None, scale=1, dof=None, allow_singular=False)``
        Probability density function.
    ``logpdf(x, location=None, scale=1, dof=None, allow_singular=False)``
        Log of the probability density function.
    ``cdf(x, location=None, scale=1, dof=None, allow_singular=False, maxpts=1000000*dim, abseps=1e-5, releps=1e-5)``
        Cumulative distribution function.
    ``logcdf(x, location=None, scale=1, dof=None, allow_singular=False, maxpts=1000000*dim, abseps=1e-5, releps=1e-5)``
        Log of the cumulative distribution function.
    ``rvs(location=None, scale=1, dof=None, size=1, random_state=None)``
        Draw random samples from a multivariate t-distribution.
    ``entropy()``
        Compute the differential entropy of the multivariate Student's.

    Parameters
    ----------
    x : array_like
        Quantiles, with the last axis of `x` denoting the components.
    %(_mvt_doc_default_callparams)s
    %(_doc_random_state)s

    Alternatively, the object may be called (as a function) to fix the mean
    and covariance parameters, returning a "frozen" multivariate Student's
    random variable:

    rv = multivariate_t(location=None, scale=1, dof=None, allow_singular=False)
        - Frozen object with the same methods but holding the given
          mean and covariance fixed.

    Notes
    -----
    %(_mvt_doc_callparams_note)s

    The scale matrix `scale` must be a (symmetric) positive
    semi-definite matrix. The determinant and inverse of `scale` are computed
    as the pseudo-determinant and pseudo-inverse, respectively, so
    that `scale` does not need to have full rank.

    The probability density function for `multivariate_t` is

    .. math::

        f(x) = \frac{\Gamma[(\nu+p)/2]}{\Gamma(\nu/2)\sqrt{(\pi \nu)^p |\Sigma|}}
               \left(1+\frac{1}{\nu} (x - \mu)^T \Sigma^{-1} (x - \mu) \right)^{-(\nu+p)/2}

    where :math:`\mu` is the location, :math:`\Sigma` the scale matrix,
    :math:`\nu` is the degrees-of-freedom, and :math:`k` is the
    dimension of the space where :math:`x` takes values.

    .. versionadded:: 0.0.1

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from student_mixture import multivariate_t

    >>> x = np.linspace(0, 5, 10, endpoint=False)
    >>> y = multivariate_t.pdf(x, location=2.5, scale=0.5, dof=30); y
    array([ 0.00108914,  0.01033349,  0.05946514,  0.20755375,  0.43939129,
            0.56418958,  0.43939129,  0.20755375,  0.05946514,  0.01033349])
    >>> fig1 = plt.figure()
    >>> ax = fig1.add_subplot(111)
    >>> ax.plot(x, y)

    The input quantiles can be any shape of array, as long as the last
    axis labels the components.  This allows us for instance to
    display the frozen pdf for a non-isotropic random variable in 2D as
    follows:

    >>> x, y = np.mgrid[-1:1:.01, -1:1:.01]
    >>> pos = np.dstack((x, y))
    >>> rv = multivariate_t([0.5, -0.2], [[2.0, 0.3], [0.3, 0.5]], 20)
    >>> fig2 = plt.figure()
    >>> ax2 = fig2.add_subplot(111)
    >>> ax2.contourf(x, y, rv.pdf(pos))

    """

    def __init__(self, seed=None):
        super(multivariate_t_gen, self).__init__(seed)
        self.__doc__ = doccer.docformat(self.__doc__, mvt_docdict_params)

    def __call__(self, location=None, scale=1, dof=None, allow_singular=False, seed=None):
        """
        Create a frozen multivariate t distribution.

        See `multivariate_t_frozen` for more information.

        """
        return multivariate_t_frozen(location, scale, dof,
                                          allow_singular=allow_singular,
                                          seed=seed)

    def _process_parameters(self, dim, location, scale, dof):
        """
        Infer dimensionality from location or scale matrix, ensure that
        location and scale are full vector resp. matrix, and ensure that
        2 < dof.

        """

        # Try to infer dimensionality
        if dim is None:
            if location is None:
                if scale is None:
                    dim = 1
                else:
                    scale = np.asarray(scale, dtype=float)
                    if scale.ndim < 2:
                        dim = 1
                    else:
                        dim = scale.shape[0]
            else:
                location = np.asarray(location, dtype=float)
                dim = location.size
        else:
            if not np.isscalar(dim):
                raise ValueError("Dimension of random variable must be "
                                 "a scalar.")

        # Check input sizes and return full arrays for location
        # and scale if necessary
        if location is None:
            location = np.zeros(dim)
        location = np.asarray(location, dtype=float)

        if scale is None:
            scale = 1.0
        scale = np.asarray(scale, dtype=float)

        if dim == 1:
            location.shape = (1,)
            scale.shape = (1, 1)

        if location.ndim != 1 or location.shape[0] != dim:
            raise ValueError("Array 'location' must be a vector of length %d." %
                             dim)
        if scale.ndim == 0:
            scale = scale * np.eye(dim)
        elif scale.ndim == 1:
            scale = np.diag(scale)
        elif scale.ndim == 2 and scale.shape != (dim, dim):
            rows, cols = scale.shape
            if rows != cols:
                msg = ("Array 'scale' must be square if it is two dimensional,"
                       " but scale.shape = %s." % str(scale.shape))
            else:
                msg = ("Dimension mismatch: array 'scale' is of shape %s,"
                       " but 'location' is a vector of length %d.")
                msg = msg % (str(scale.shape), len(location))
            raise ValueError(msg)
        elif scale.ndim > 2:
            raise ValueError("Array 'scale' must be at most two-dimensional,"
                             " but scale.ndim = %d" % scale.ndim)
        if dof is None:
            dof = np.inf
        elif dof > 2:
            pass
        else:
            raise ValueError("Parameter 'dof' must be >2,"
                             " but dof = %d" % dof)

        return dim, location, scale, dof

    def _process_quantiles(self, x, dim):
        """
        Adjust quantiles array so that last axis labels the components of
        each data point.

        """
        x = np.asarray(x, dtype=float)

        if x.ndim == 0:
            x = x[np.newaxis]
        elif x.ndim == 1:
            if dim == 1:
                x = x[:, np.newaxis]
            else:
                x = x[np.newaxis, :]

        return x

    def _logpdf(self, x, location, prec_U, log_det_scale, rank, dof):
        """
        Parameters
        ----------
        x : ndarray
            Points at which to evaluate the log of the probability
            density function
        location : ndarray
            Location of the distribution
        prec_U : ndarray
            A decomposition such that np.dot(prec_U, prec_U.T)
            is the precision matrix, i.e. inverse of the scale matrix.
        log_det_cov : float
            Logarithm of the determinant of the scale matrix
        rank : int
            Rank of the scale matrix.
        dof : scalar
            Degrees-of-freedom of the distribution

        Notes
        -----
        As this function does no argument checking, it should not be
        called directly; use 'logpdf' instead.

        """
        if dof == np.inf:
            return multivariate_normal._logpdf(x, location, prec_U, log_det_scale, rank)
        else:
            gamma_ratio = gammaln((dof + rank) / 2) - gammaln(dof / 2)
            if gamma_ratio != np.nan:
                dev = x - location
                log_maha = np.log(1 + np.sum(np.square(np.dot(dev, prec_U)), axis=-1) / dof)
                return gamma_ratio - 0.5 * (rank * (_LOG_PI + np.log(dof)) + log_det_scale + (dof + rank) * log_maha)
            else:
                return multivariate_normal._logpdf(x, location, prec_U, log_det_scale, rank)

    def logpdf(self, x, location=None, scale=1, dof=None, allow_singular=False):
        """
        Log of the multivariate Student's t probability density function.

        Parameters
        ----------
        x : array_like
            Quantiles, with the last axis of `x` denoting the components.
        %(_mvt_doc_default_callparams)s

        Returns
        -------
        pdf : ndarray or scalar
            Log of the probability density function evaluated at `x`

        Notes
        -----
        %(_mvt_doc_callparams_note)s

        """
        dim, location, scale, dof = self._process_parameters(None, location, scale, dof)
        x = self._process_quantiles(x, dim)
        psd = _PSD(scale, allow_singular=allow_singular)
        out = self._logpdf(x, location, psd.U, psd.log_pdet, psd.rank, dof)
        return _squeeze_output(out)

    def pdf(self, x, location=None, scale=1, dof=None, allow_singular=False):
        """
        Multivariate Student's t probability density function.

        Parameters
        ----------
        x : array_like
            Quantiles, with the last axis of `x` denoting the components.
        %(_mvt_doc_default_callparams)s

        Returns
        -------
        pdf : ndarray or scalar
            Probability density function evaluated at `x`

        Notes
        -----
        %(_mvt_doc_callparams_note)s

        """
        dim, location, scale, dof = self._process_parameters(None, location, scale, dof)
        x = self._process_quantiles(x, dim)
        psd = _PSD(scale, allow_singular=allow_singular)
        out = np.exp(self._logpdf(x, location, psd.U, psd.log_pdet, psd.rank, dof))
        return _squeeze_output(out)

    def _cdf(self, x, location, scale, dof, maxpts, abseps, releps):
        """
        Parameters
        ----------
        x : ndarray
            Points at which to evaluate the cumulative distribution function.
        location : ndarray
            Location of the distribution
        scale : array_like
            Scale matrix of the distribution
        dof : scalar
            Degrees-of-freedom of the distribution
        maxpts: integer
            The maximum number of points to use for integration (only used when dim > 3)
        abseps: float
            Absolute error tolerance (only used when dim > 1)
        releps: float
            Relative error tolerance (only used when dim == 2 or dim == 3)

        Notes
        -----
        As this function does no argument checking, it should not be
        called directly; use 'cdf' instead.

        .. versionadded:: 1.0.0

        """
        if dof == np.inf:
            return multivariate_normal.cdf(x, location, scale, maxpts=maxpts, abseps=abseps, releps=releps)
        else:
            return _multivariate_t_cdf(x, location, scale, dof, maxpts=maxpts, abseps=abseps, releps=releps)

    def logcdf(self, x, location=None, scale=1, dof=None, allow_singular=False, maxpts=None,
               abseps=1e-5, releps=1e-5):
        """
        Log of the multivariate Student's t cumulative distribution function.

        Parameters
        ----------
        x : array_like
            Quantiles, with the last axis of `x` denoting the components.
        %(_mvt_doc_default_callparams)s
        maxpts: integer, optional
            The maximum number of points to use for integration
            (default `1000000*dim`)
        abseps: float, optional
            Absolute error tolerance (default 1e-5)
        releps: float, optional
            Relative error tolerance (default 1e-5)

        Returns
        -------
        cdf : ndarray or scalar
            Log of the cumulative distribution function evaluated at `x`

        Notes
        -----
        %(_mvt_doc_callparams_note)s

        .. versionadded:: 1.0.0

        """
        dim, location, scale, dof = self._process_parameters(None, location, scale, dof)
        x = self._process_quantiles(x, dim)
        # Use _PSD to check covariance matrix
        _PSD(scale, allow_singular=allow_singular)
        if not maxpts:
            maxpts = 1000000 * dim
        out = np.log(self._cdf(x, location, scale, dof, maxpts, abseps, releps))
        return out

    def cdf(self, x, location=None, scale=1, dof=None, allow_singular=False, maxpts=None,
            abseps=1e-5, releps=1e-5):
        """
        Multivariate Student's t cumulative distribution function.

        Parameters
        ----------
        x : array_like
            Quantiles, with the last axis of `x` denoting the components.
        %(_mvt_doc_default_callparams)s
        maxpts: integer, optional
            The maximum number of points to use for integration
            (default `1000000*dim`)
        abseps: float, optional
            Absolute error tolerance (default 1e-5)
        releps: float, optional
            Relative error tolerance (default 1e-5)

        Returns
        -------
        cdf : ndarray or scalar
            Cumulative distribution function evaluated at `x`

        Notes
        -----
        %(_mvt_doc_callparams_note)s

        .. versionadded:: 1.0.0

        """
        dim, location, scale, dof = self._process_parameters(None, location, scale, dof)
        x = self._process_quantiles(x, dim)
        # Use _PSD to check covariance matrix
        _PSD(scale, allow_singular=allow_singular)
        if not maxpts:
            maxpts = 1000000 * dim
        out = self._cdf(x, location, scale, dof, maxpts, abseps, releps)
        return out

    def rvs(self, location=None, scale=1, dof=None, size=1, random_state=None):
        """
        Draw random samples from a multivariate Student's t distribution.

        Parameters
        ----------
        %(_mvt_doc_default_callparams)s
        size : integer, optional
            Number of samples to draw (default 1).
        %(_doc_random_state)s

        Returns
        -------
        rvs : ndarray or scalar
            Random variates of size (`size`, `N`), where `N` is the
            dimension of the random variable.

        Notes
        -----
        %(_mvt_doc_callparams_note)s

        """
        _, location, scale, dof = self._process_parameters(None, location, scale, dof)
        if dof == np.inf:
            random_state = self._get_random_state(random_state)
            out = random_state.multivariate_normal(location, scale, size)
            return _squeeze_output(out)
        else:
            return _multivariate_t_random(location, scale, dof, size, random_state)

    def entropy(self, location=None, scale=1, dof=None):
        """
        Compute the differential entropy of the multivariate Student's t.

        Parameters
        ----------
        %(_mvt_doc_default_callparams)s

        Returns
        -------
        h : scalar
            Entropy of the multivariate Student's t distribution

        Notes
        -----
        %(_mvt_doc_callparams_note)s

        """
        if dof == np.inf:
            return multivariate_normal.entropy(location, scale)
        else:
            dim, location, scale, dof = self._process_parameters(None, location, scale, dof)
            _, logdet = np.linalg.slogdet(scale)
            loggamma_ratio = gammaln(0.5 * dof) - gammaln(0.5 * (dof + dim)) + (0.5 * dim) * (np.log(dof) + _LOG_PI)
            digamma_part = (0.5 * (dof + dim)) * (digamma(0.5 * (dof + dim)) - digamma(0.5 * dof))
            return 0.5 * logdet + loggamma_ratio + digamma_part


multivariate_t = multivariate_t_gen()


class multivariate_t_frozen(multi_rv_frozen):
    def __init__(self, location=None, scale=1, dof=None, allow_singular=False, seed=None,
                 maxpts=None, abseps=1e-5, releps=1e-5):
        """
        Create a frozen multivariate Student's t-distribution.

        Parameters
        ----------
        location : array_like, optional
            Location of the distribution (default zero)
        scale : array_like, optional
            Scale matrix of the distribution (default one)
        dof : None or scalar, optional
            Degrees-of-freedom of the distribution (default numpy.inf)
        allow_singular : bool, optional
            If this flag is True then tolerate a singular
            scale matrix (default False).
        seed : None or int or np.random.RandomState instance, optional
            This parameter defines the RandomState object to use for drawing
            random variates.
            If None (or np.random), the global np.random state is used.
            If integer, it is used to seed the local RandomState instance
            Default is None.
        maxpts: integer, optional
            The maximum number of points to use for integration of the
            cumulative distribution function (default `1000000*dim`)
        abseps: float, optional
            Absolute error tolerance for the cumulative distribution function
            (default 1e-5)
        releps: float, optional
            Relative error tolerance for the cumulative distribution function
            (default 1e-5)

        Examples
        --------
        When called with the default parameters, this will create a 1D random
        variable with mean 0 and covariance 1:

        >>> from student_mixture import multivariate_t
        >>> r = multivariate_t()
        >>> r.location
        array([ 0.])
        >>> r.scale
        array([[1.]])
        >>> r.dof
        inf

        """
        self._dist = multivariate_t_gen(seed)
        self.dim, self.location, self.scale, self.dof = self._dist._process_parameters(
                                                            None, location, scale, dof)
        self.scale_info = _PSD(self.scale, allow_singular=allow_singular)
        if not maxpts:
            maxpts = 1000000 * self.dim
        self.maxpts = maxpts
        self.abseps = abseps
        self.releps = releps

    def logpdf(self, x):
        x = self._dist._process_quantiles(x, self.dim)
        out = self._dist._logpdf(x, self.location, self.scale_info.U,
                                 self.scale_info.log_pdet, self.scale_info.rank, self.dof)
        return _squeeze_output(out)

    def pdf(self, x):
        return np.exp(self.logpdf(x))

    def logcdf(self, x):
        return np.log(self.cdf(x))

    def cdf(self, x):
        x = self._dist._process_quantiles(x, self.dim)
        out = self._dist._cdf(x, self.location, self.scale, self.dof, self.maxpts, self.abseps,
                              self.releps)
        return _squeeze_output(out)

    def rvs(self, size=1, random_state=None):
        return self._dist.rvs(self.location, self.scale, self.dof, size, random_state)

    def entropy(self):
        """
        Computes the differential entropy of the multivariate Student's t.

        Returns
        -------
        h : scalar
            Entropy of the multivariate Student's t distribution

        """
        if self.dof == np.inf:
            return multivariate_normal.entropy(self.location, self.scale)
        else:
            return self._dist.entropy(self.location, self.scale, self.dof)


# Set frozen generator docstrings from corresponding docstrings in
# multivariate_t_gen and fill in default strings in class docstrings
for name in ['logpdf', 'pdf', 'logcdf', 'cdf', 'rvs']:
    method = multivariate_t_gen.__dict__[name]
    method_frozen = multivariate_t_frozen.__dict__[name]
    method_frozen.__doc__ = doccer.docformat(method.__doc__,
                                             mvt_docdict_noparams)
    method.__doc__ = doccer.docformat(method.__doc__, mvt_docdict_params)



"""Student's t-distribution Fitting methods"""

# Author: Omri Tomer <omritomer1@mail.tau.ac.il>
# License: BSD 3 clause

import numpy as np

from scipy import linalg
from scipy.special import gammaln, digamma, polygamma
from scipy.optimize import newton
from scipy.stats import chi2, multivariate_normal as mvn
from sklearn.utils.extmath import row_norms
from sklearn.utils import check_array

import warnings

###############################################################################
# Functions to be used by the MultivariateTFit class

def _check_X(X, n_features=None, ensure_min_samples=1):
    """Check the input data X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    Returns
    -------
    X : array, shape (n_samples, n_features)
    """
    X = check_array(X, dtype=[np.float64, np.float32],
                    ensure_min_samples=ensure_min_samples)

    if n_features is not None and X.shape[1] != n_features:
        raise ValueError("Expected the input data X have %d features, "
                         "but got %d features"
                         % (n_features, X.shape[1]))
    return X


def _check_shape(param, param_shape, name):
    """Validate the shape of the input parameter 'param'.

    Parameters
    ----------
    param : array

    param_shape : tuple

    name : string
    """
    param = np.array(param)
    if param.shape != param_shape:
        raise ValueError("The parameter '%s' should have the shape of %s, "
                         "but got %s" % (name, param_shape, param.shape))


def _check_location(location, n_features):
    """Validate the provided 'means'.

    Parameters
    ----------
    mean : array-like, shape (n_components, n_features)
        The centers of the current components.

    n_components : int
        Number of components.

    n_features : int
        Number of features.

    Returns
    -------
    means : array, (n_components, n_features)
    """
    location = check_array(location, dtype=[np.float64, np.float32], ensure_2d=False)
    _check_shape(location, (n_features,), 'location')
    return location


def _check_dof(dof):
    """Check the user provided 'dofs'.

    Parameters
    ----------
    dofs : array-like, shape (n_components,)
        The degrees-of-freedom of each mixture.

    n_components : int
        Number of components.

    Returns
    -------
    weights : array, shape (n_components,)
    """
    # check range
    if dof < 2:
        raise ValueError("The parameter 'dof' should be in the range "
                         "[2, inf), but got min value %.5f"
                         % dof)
    return dof


def _check_precision_positivity(precision, scale_type):
    """Check a precision vector is positive-definite."""
    if np.any(np.less_equal(precision, 0.0)):
        raise ValueError("'%s precision' should be "
                         "positive" % scale_type)


def _check_precision_matrix(precision, scale_type):
    """Check a precision matrix is symmetric and positive-definite."""
    if not (np.allclose(precision, precision.T) and
            np.all(linalg.eigvalsh(precision) > 0.)):
        raise ValueError("'%s precision' should be symmetric, "
                         "positive-definite" % scale_type)


def _check_precision_full(precision, scale_type):
    """Check the precision matrices are symmetric and positive-definite."""
    _check_precision_matrix(precision, scale_type)


def _check_precision(precision, scale_type, n_features):
    """Validate user provided precisions.

    Parameters
    ----------
    precisions : array-like
        'full' : shape of (n_features, n_features)
        'diag' : shape of (n_features,)
        'spherical' : scalar

    scale_type : string

    n_features : int
        Number of features.

    Returns
    -------
    precisions : array
    """
    precision = check_array(precision, dtype=[np.float64, np.float32],
                            ensure_2d=False,
                            allow_nd=False)

    precision_shape = {'full': (n_features, n_features),
                       'diag': (n_features,),
                       'spherical': (1,)}
    _check_shape(precision, precision_shape[scale_type],
                 '%s precision' % scale_type)

    _check_precision = {'full': _check_precision_matrix,
                        'diag': _check_precision_positivity,
                        'spherical': _check_precision_positivity}
    _check_precision[scale_type](precision, scale_type)
    return precision


###############################################################################
# Student mixture parameters estimators (used by the M-Step)

def _estimate_student_scales_full(gamma_priors, X, location, reg_scale):
    """Estimate the full scale matrix.

    Parameters
    ----------
    gamma_priors : array-like, shape (n_samples,)

    X : array-like, shape (n_samples, n_features)

    location : array-like, shape (n_features,)

    reg_scale : float

    Returns
    -------
    scales : array, shape (n_features, n_features)
        The scale matrix of the distribution.
    """
    n_samples, _ = X.shape
    diff = X - location
    scale = np.dot((gamma_priors[:, np.newaxis] * diff).T, diff) / n_samples
    return scale + reg_scale


def _estimate_student_scales_diag(gamma_priors, X, location, reg_scale):
    """Estimate the diagonal scale vector.

    Parameters
    ----------
    gamma_priors : array-like, shape (n_samples,)

    X : array-like, shape (n_samples, n_features)

    location : array-like, shape (n_features,)

    reg_scale : float

    Returns
    -------
    scales : array, shape (n_features,)
        The scale matrix of the distribution.
    """
    n_samples = X.shape[0]
    avg_X2 = np.dot(gamma_priors, X * X)
    avg_location2 = gamma_priors.sum() * (location * location)
    avg_X_location = location * np.dot(gamma_priors, X)
    return (avg_X2 - 2 * avg_X_location + avg_location2) / n_samples + reg_scale


def _estimate_student_scales_spherical(gamma_priors, X, location, reg_scale):
    """Estimate the spherical scale value.

    Parameters
    ----------
    gamma_priors : array-like, shape (n_samples,)

    X : array-like, shape (n_samples, n_features)

    location : array-like, shape (n_features,)

    reg_scale : float

    Returns
    -------
    scales : scalar
        The scale matrix of the distribution.
    """
    return _estimate_student_scales_diag(gamma_priors, X,
                                         location, reg_scale).mean()


def _estimate_student_parameters(X, gamma_priors, reg_scale, scale_type):
    """Estimate the Student's t-distribution parameters.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        The input data array.

    gamma_priors : array-like, shape (n_samples,)
        The gamma weights for each data sample in X.

    reg_scale : float
        The regularization added to the diagonal of the scale matrix.

    scale_type : {'full', 'diag', 'spherical'}
        The type of scale matrix.

    Returns
    -------
    location : array-like, shape (n_features,)
        The centers of the distribution.

    scale : array-like
        The scale matrix of the distribution.
        The shape depends of the scale_type.
    """
    location = np.dot((gamma_priors).T, X) / (gamma_priors).T.sum()
    scale = {"full": _estimate_student_scales_full,
             "diag": _estimate_student_scales_diag,
             "spherical": _estimate_student_scales_spherical
             }[scale_type](gamma_priors, X, location, reg_scale)
    return location, scale


def _compute_precision_cholesky(scale, scale_type):
    """Compute the Cholesky decomposition of the precision matrix.

    Parameters
    ----------
    scale : array-like
        The scale matrix of the distribution.
        The shape depends of the scale_type.

    scale_type : {'full', 'diag', 'spherical'}
        The type of scale matrix.

    Returns
    -------
    precision_cholesky : array-like
        The cholesky decomposition of the precision matrix of the
        distribution. The shape depends of the scale_type.
    """
    estimate_precision_error_message = (
        "Fitting the t-distribution failed because it has an ill-defined "
        "empirical scale (for instance caused by singleton or collapsed"
        "samples). Try to increase reg_scale.")

    if scale_type == 'full':
        try:
            cov_chol = linalg.cholesky(scale, lower=True)
        except linalg.LinAlgError:
            raise ValueError(estimate_precision_error_message)
        n_features = scale.shape[0]
        precision_chol = linalg.solve_triangular(cov_chol,
                                                     np.eye(n_features),
                                                     lower=True).T
    else:
        if np.any(np.less_equal(scale, 0.0)):
            print(scale)
            raise ValueError(estimate_precision_error_message)
        precision_chol = 1. / np.sqrt(scale)
    return precision_chol


def _compute_gamma_priors(X, location, precision_chol, scale_type, dof):
    """Estimate the gamma weights of samples.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    location : array-like, shape (n_features,)

    precision_chol : array-like
        Cholesky decompositions of the precision matrix.
        'full' : shape of (n_features, n_features)
        'diag' : shape of (n_features,)
        'spherical' : scalar

    scale_type : {'full', 'diag', 'spherical'}

    dofs : scalar

    Returns
    -------
    gamma_weights : array, shape (n_samples,)
    """

    _, n_features = X.shape

    if scale_type == 'full':
        y = np.dot(X, precision_chol) - np.dot(location, precision_chol)
        prob = np.sum(np.square(y), axis=1)

    elif scale_type == 'diag':
        precision = precision_chol ** 2
        prob = (np.sum((location ** 2 * precision)) -
                2. * np.dot(X, (location * precision).T) +
                np.dot(X ** 2, precision))

    else:
        precision = precision_chol ** 2
        prob = (np.sum(location ** 2) * precision -
                2 * np.dot(X, location.T * precision) +
                precision * row_norms(X, squared=True))

    return (dof + n_features) / (dof + prob)


def _compute_delta(X, location, precision_chol, scale_type):
    """Compute the delta (distance from center) of samples.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    location : array-like, shape (n_features,)

    precisions_chol : array-like
        Cholesky decompositions of the precision matrix.
        'full' : shape of (n_features, n_features)
        'diag' : shape of (n_features,)
        'spherical' : scalar

    scale_type : {'full', 'diag', 'spherical'}

    Returns
    -------
    delta : array, shape (n_samples,)
    """   
    if scale_type == 'full':
        y = np.dot(X, precision_chol) - np.dot(location, precision_chol)
        return np.sum(np.square(y), axis=1)

    elif scale_type == 'diag':
        precision = precision_chol ** 2
        return (np.sum((location ** 2 * precision)) -
                2. * np.dot(X, (location * precision).T) +
                np.dot(X ** 2, precision))

    else:
        precision = precision_chol ** 2
        return (np.sum(location ** 2) * precision -
                2 * np.dot(X, location.T * precision) +
                precision * row_norms(X, squared=True))


###############################################################################
# Student mixture probability estimators
def _compute_log_det_cholesky(matrix_chol, scale_type, n_features):
    """Compute the log-det of the cholesky decomposition of matrix.

    Parameters
    ----------
    matrix_chol : array-like
        Cholesky decompositions of the matrices.
        'full' : shape of (n_features, n_features)
        'diag' : shape of (n_features,)
        'spherical' : scalar

    scale_type : {'full', 'diag', 'spherical'}

    n_features : int
        Number of features.

    Returns
    -------
    log_det_precision_chol : array-like, shape (n_components,)
        The determinant of the precision matrix.
    """
    if scale_type == 'full':
        log_det_chol = (np.sum(np.log(np.diag(matrix_chol))))

    elif scale_type == 'diag':
        log_det_chol = (np.sum(np.log(matrix_chol)))

    else:
        log_det_chol = n_features * (np.log(matrix_chol))

    return log_det_chol


def _estimate_log_student_prob(X, location, precision_chol, scale_type, dof):
    """Estimate the log Student probability.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    location : array-like, shape (n_features,)

    precisions_chol : array-like
        Cholesky decompositions of the precision matrix.
        'full' : shape of (n_features, n_features)
        'diag' : shape of (n_features,)
        'spherical' : scalar

    scale_type : {'full', 'diag', 'spherical'}

    dof : scalar

    Returns
    -------
    log_prob : array, shape (n_samples,)
    """
    _, n_features = X.shape
    # det(precision_chol) is half of det(precision)
    log_det = _compute_log_det_cholesky(precision_chol, scale_type, n_features)

    if scale_type == 'full':
        y = np.dot(X, precision_chol) - np.dot(location, precision_chol)
        prob = np.sum(np.square(y), axis=1)

    elif scale_type == 'diag':
        precision = precision_chol ** 2
        prob = (np.sum((location ** 2 * precision)) -
                2. * np.dot(X, (location * precision).T) +
                np.dot(X ** 2, precision.T))

    else:
        precision = precision_chol ** 2
        prob = (np.sum(location ** 2) * precision -
                2 * np.dot(X, location * precision) +
                precision * row_norms(X, squared=True))
    log_prob = np.log(1 + (prob / dof))
    return gammaln(.5 * (dof + n_features)) - gammaln(.5 * dof) - .5 * (n_features * np.log(dof * np.pi) +
                                                                        (dof + n_features) * log_prob) + log_det


def _initialize_dof(X, location, precision_chol, scale_type, max_iter=100, tol=1e-3):
    """Initial degrees-of-freedom estimation

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)

    location : array-like, shape (n_features,)

    precisions_chol : array-like
        Cholesky decompositions of the precision matrix.
        'full' : shape of (n_features, n_features)
        'diag' : shape of (n_features,)
        'spherical' : scalar

    scale_type : {'full', 'diag', 'spherical'}

    max_iter : int, defaults to 100.
        The number of iterations to perform for degrees-of-freedom estimation.

    tol : float, defaults to 1e-3.
        The degrees-of-freedom estimation convergence threshold.

    Returns
    -------
    dof : scalar
    """

    n_samples, n_features = X.shape
    delta = _compute_delta(X, location, precision_chol, scale_type)
    dof = 30
    def init_function(df):
        sample_dependent_component = -(np.log(df + delta) + 1.0 * (df + n_features) / (df + delta))
        gamma_component = digamma(1.0 * (df + n_features) / 2.0) - digamma(1.0 * df / 2.0)
        log_component = 1.0 * np.log(df) + 1.0
        return gamma_component + log_component + (1.0 / n_samples) * (
            np.sum(sample_dependent_component)
        )

    def init_first_derivative(df):
        sample_dependent_component = -(1.0 / (df + delta) + 1.0 * (n_features - delta) / ((df + delta) ** 2))
        gamma_component = (1.0 / 2.0) * (polygamma(1, 1.0 * (df + n_features) / 2.0) - polygamma(1, 1.0 * df / 2.0))
        log_component = 1.0 / df
        return gamma_component + log_component - (1 / n_samples) * (
            np.sum(sample_dependent_component)
        )

    def init_second_derivative(df):
        sample_dependent_component = 1.0 / ((df + delta) ** 2) + 2.0 * (n_features - delta) / ((df + delta) ** 3)
        gamma_component = (1.0 / 4.0) * (polygamma(2.0, 1.0 * (df + n_features) / 2.0) - polygamma(2, 1.0 * df / 2.0))
        log_component = -1.0 / (df ** 2)
        return gamma_component + log_component + (1.0 / n_samples) * (
            np.sum(sample_dependent_component)
        )
    for i in range(3, 31):
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                new_dof = newton(
                    init_function,
                    i,
                    init_first_derivative,
                    args=(),
                    maxiter=max_iter,
                    tol=tol,
                    fprime2=init_second_derivative,
                    full_output=False,
                    disp=False,
                )
                if np.isfinite(new_dof) and new_dof > 2:
                    return new_dof

        except Exception:
            pass
    return dof


def _estimate_dof(gamma_priors, current_dof, n_features, max_iter=100, tol=1e-3):
    """Estimate the degrees of freedom

    Parameters
    ----------
    gamma_priors : array-like, shape (n_samples,)
        The gamma priors for each data sample in X.

    current_dof : scalar
        Current degrees-of-freedom estimation.

   n_features : int
        Number of features.

    max_iter : int, defaults to 100.
        The number of iterations to perform for degrees-of-freedom estimation.

    tol : float, defaults to 1e-3.
        The degrees-of-freedom estimation convergence threshold.

    Returns
    -------
    dof : scalar
    """
    n_samples = gamma_priors.shape[0]
    constant = (
        1.0
        + (1.0 / n_samples)
        * (np.log(gamma_priors) - gamma_priors).sum()
        + digamma(1.0 * (current_dof + n_features) / 2.0)
        - np.log(1.0 * (current_dof + n_features) / 2.0)
    )

    def function(df):
        return -digamma(1.0 * df / 2.0) + np.log(1.0 * df / 2.0) + constant

    def first_derivative(df):
        return -(1.0 / 2.0) * polygamma(1, 1.0 * df / 2.0) + 1.0 / df

    def second_derivative(df):
        return -(1.0 / 4.0) * polygamma(2, 1.0 * df / 2.0) - 1.0 / (df * df)

    dof = newton(
        function,
        current_dof,
        first_derivative,
        args=(),
        maxiter=max_iter,
        tol=tol,
        fprime2=second_derivative,
        full_output=False,
        disp=False,
    )

    return dof



import numpy as np
from scipy.special import erfc, erfcinv, gammaincinv


def normp(z):
    return 0.5 * erfc(-z / (2 ** 0.5))


def normq(z):
    return -(2 ** 0.5) * erfcinv(2 * z)


def chiq(z, nu):
    return (2 * gammaincinv(nu / 2, z)) ** (0.5)


def _multivariate_t_cdf_single_sample(x, chol, dof, tol=1e-4, max_evaluations=1e+7):
    """
    Wrapper function for estimating the cdf of single sample 'x'.

    Parameters
    ----------
    x : array_like
        Standardized sample, shape (n_features,)
    chol : array_like, shape (n_features, n_features)
        Cholesky decomposition of the correlation matrix of the distribution
    dof : float
        Degrees-of-freedom of the distribution, must be a positive real number
    tol: float
        Tolerance for quasi-Monte Carlo algorithm
    max_evaluations: float
        Maximum points to evaluate with quasi-Monte Carlo algorithm
    

    Returns
    -------
    cdf : scalar
        Cumulative density function evaluated at `x`
    """
    if np.all(x == np.inf):
        return 1, 0

    elif np.all(x == -np.inf):
        return 0, 0

    else:
        return _multivariate_t_cdf_quasi_monte_carlo(x, chol, dof, tol=tol, max_evaluations=max_evaluations)


def _multivariate_t_cdf_quasi_monte_carlo(x, chol, dof, tol=1e-4, max_evaluations=1e+7):
    """
    Function for estimating the cdf of single sample 'x'.
    Adapted from MATLAB's mvtcdfqmc function.
    Algorithm based on:
    Genz, A., & Bretz, F. (1999). Numerical computation of multivariate
    t-probabilities with application to power calculation of multiple
    contrasts. Journal of Statistical Computation and Simulation, 63(4),
    103-117.
    Genz, A., & Bretz, F. (2002). Comparison of methods for the computation
    of multivariate t probabilities. Journal of Computational and Graphical
    Statistics, 11(4), 950-971.

    Parameters
    ----------
    x : array_like
        Standardized sample, shape (n_features,)
    chol : array_like, shape (n_features, n_features)
        Cholesky decomposition of the correlation matrix of the distribution
    dof : float
        Degrees-of-freedom of the distribution, must be a positive real number
    tol: float
        Tolerance for quasi-Monte Carlo algorithm
    max_evaluations: float
        Maximum points to evaluate with quasi-Monte Carlo algorithm
    

    Returns
    -------
    cdf : scalar
        Cumulative density function evaluated at `x`
    """
    primes = [173, 263, 397, 593, 907, 1361, 2053, 3079, 4621, 6947, 10427, 15641, 23473,
              35221, 52837, 79259, 118891, 178349, 267523, 401287, 601942, 902933, 1354471, 2031713]
    m = chol.shape[0]
    c = np.diag(chol)
    x = x / c
    chol = chol / np.tile(c, (m, 1))

    mc_repetitions = 25
    p = 0
    sigma_squared = np.inf

    fun_evals = 0
    err = np.nan

    for i in primes:
        if (fun_evals + 2 * mc_repetitions * i) > max_evaluations:
            break

        q = 2 ** (np.arange(1, m + 1) / (m + 1))

        tau_hat, sigma_squared_tau_hat = _mvt_qmc(x, chol, dof, mc_repetitions, i, q)

        fun_evals += 2 * mc_repetitions * i

        p = p + (tau_hat - p) / (1 + sigma_squared_tau_hat / sigma_squared)
        sigma_squared = sigma_squared_tau_hat / (1 + sigma_squared_tau_hat / sigma_squared)
        err = 3.5 * np.sqrt(sigma_squared)

        if err < tol:
            return p, err
    return p, err


def _mvt_qmc(x, chol, dof, mc_repetitions, prime, q):
    qq = np.outer(np.arange(1, prime + 1), q)
    tau_hat = np.zeros(mc_repetitions)

    for rep in range(mc_repetitions):
        r = np.tile(np.random.uniform(low=0.0, high=1.0, size=len(q)), (prime, 1))
        w = abs(2 * ((qq + r) % 1) - 1)
        tau_hat[rep] = (_f_qrsvn(x, chol, dof, w) + _f_qrsvn(x, chol, dof, 1 - w)) / 2

    return np.mean(tau_hat), np.var(tau_hat) / mc_repetitions


def _f_qrsvn(x, chol, dof, w):
    n_points = w.shape[0]
    dim = x.shape[0]
    eps = np.finfo(float).eps

    s_nu = chiq(w[:, -1], dof) / (dof ** 0.5)
    emd = normp(s_nu * x[0])
    tau = np.copy(emd)
    y = np.zeros((n_points, dim))

    for i in range(1, dim):
        z = np.minimum(np.maximum(emd * w[:, -1 - i], eps / 2), 1 - eps / 2)
        y[:, i - 1] = normq(z)
        y_sum = np.dot(y, chol[:, i])
        emd = normp(s_nu * x[i] - y_sum)
        tau = tau * emd

    return tau.sum() / tau.shape[0]



from __future__ import division, print_function, absolute_import

import numpy as np

from scipy.stats import t
from scipy.stats._multivariate import multivariate_normal
from scipy.stats._continuous_distns import chi2
from .multivariate_t_cdf import (_standard_t_cdf_univariate, _standard_t_cdf_bivariate, _standard_t_cdf_multivariate)

###############################################################################
# Multivariate Student's t-distribution functions

def _multivariate_t_random(location, scale, dof, n_samples, random_state):
    """
    Generate random samples from the multivariate Student's t-distribution

    Parameters
    ----------
    location : array_like, shape (n_features,) or (n_features,1)
        Location parameter of the distribution
        If None, the loc defined in the class initialization is used
    scale : array_like, shape (n_features, n_features)
        Scale matrix of the distribution, must be symmetric and positive definite
    dof : float
        Degrees-of-freedom of the distribution, must be a positive real number
    n_samples : int
        Number of samples to generate, must be a positive integer
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the
        random number generator; If None, the random number generator is the RandomState instance used by numpy.random.

    Returns
    -------
    X : array_like, shape (num_of_samples, num_of_variables)
        A random sample drawn from the distribution, with rows as samples and columns as variables

    """
    n_features = location.shape[0]
    x = chi2.rvs(dof, size=n_samples, random_state=random_state) / dof
    z = multivariate_normal.rvs(
        np.zeros(n_features), scale, (n_samples,), random_state
    )
    return location.T + z / np.sqrt(x)[:, None]


def _multivariate_t_cdf(x, location, scale, dof, maxpts=1e+7, abseps=1e-6, releps=1e-6):
    """
    Multivariate Student's t cumulative density function.

    Parameters
    ----------
    x : array_like
        Sample, shape (n_samples, n_features)
    location : array_like, shape (n_features,)
        Location parameter of the distribution
    scale : array_like, shape (n_features, n_features)
        Scale matrix of the distribution, must be symmetric and positive definite
    dof : float
        Degrees-of-freedom of the distribution, must be a positive real number
    maxpts: integer
        The maximum number of points to use for integration (used when n_features > 3)
    abseps: float
        Absolute error tolerance (used when n_features > 1)
    releps: float
        Relative error tolerance (used when n_features == 2 or n_features == 3)
    

    Returns
    -------
    cdf : ndarray or scalar
        Cumulative density function evaluated at `x`

    """
    if dof == np.inf:
        return multivariate_normal.cdf(x, mean=location, cov=scale, maxpts=maxpts, abseps=abseps,releps=releps)

    dim = x.shape[1]
    if x.shape[1] == 1:
        return t.cdf(x, dof, loc=location, scale=np.sqrt(scale))

    diag_scale = np.diag(scale) if dim > 1 else scale
    inv_diag_scale = 1 / diag_scale
    sqrt_inv_diag_scale = np.sqrt(inv_diag_scale)
    y = np.dot(x - location, np.diag(sqrt_inv_diag_scale)) if dim > 1 else (x - location) * sqrt_inv_diag_scale
    corr_mat = scale * np.outer(sqrt_inv_diag_scale, sqrt_inv_diag_scale) if dim > 1 else None
    if x.shape[1] == 1:
        f_cdf = _standard_t_cdf_univariate(y, dof, abseps=abseps, releps=releps)
    elif x.shape[1] == 2:
        f_cdf = _standard_t_cdf_bivariate(y, corr_mat, dof, abseps=abseps, releps=releps)
    else:
        f_cdf = _standard_t_cdf_multivariate(y, corr_mat, dof, tol=abseps, max_evaluations=maxpts)

    f_cdf[f_cdf < 0] = 0
    f_cdf[f_cdf > 1] = 1

    return f_cdf
